import warnings
import networkx as nx
import numpy as np
import scipy

warnings.filterwarnings("ignore")

"""
Reconstructing [1]
[1] Understanding Regularized Spectral Clustering via Graph Conductance Yilin Zhang, Karl Rohe: NIPS'18
https://arxiv.org/pdf/1806.01468.pdf
"""


def __regularized_laplacian_matrix(adj_matrix, tau):
    """
    Using ARPACK solver, compute the first K eigen vector.
    The laplacian is computed using the regularised formula from [2]
    [2]Kamalika Chaudhuri, Fan Chung, and Alexander Tsiatas 2018.
        Spectral clustering of graphs with general degrees in the extended planted partition model.
    L = I - D^-1/2 * A * D ^-1/2
    :param adj_matrix: adjacency matrix representation of graph where [m][n] >0 if there is edge and [m][n] = weight
    :param tau: the regularisation constant
    :return: the first K eigenvector
    """
    import scipy.sparse

    # Code inspired from nx.normalized_laplacian_matrix, with changes to allow regularisation
    n, m = adj_matrix.shape
    I = np.eye(n, m)
    diags = adj_matrix.sum(axis=1).flatten()
    # add tau to the diags to produce a regularised diags
    if tau != 0:
        diags = np.add(diags, tau)

    # diags will be zero at points where there is no edge and/or the node you are at
    #  ignore the error and make it zero later
    with scipy.errstate(divide="ignore"):
        diags_sqrt = 1.0 / scipy.sqrt(diags)
    diags_sqrt[scipy.isinf(diags_sqrt)] = 0
    D = scipy.sparse.spdiags(diags_sqrt, [0], m, n, format="csr")

    L = I - (D.dot(adj_matrix.dot(D)))
    return L


def __eigen_solver(laplacian, n_clusters):
    """
    ARPACK eigen solver in Shift-Invert Mode based on http://docs.scipy.org/doc/scipy/reference/tutorial/arpack.html
    """
    from scipy.sparse.linalg import eigsh

    lap = laplacian * -1
    v0 = np.random.uniform(-1, 1, lap.shape[0])
    eigen_values, eigen_vectors = eigsh(lap, k=n_clusters, sigma=1.0, v0=v0)
    eigen_vectors = eigen_vectors.T[n_clusters::-1]
    return eigen_values, eigen_vectors[:n_clusters].T


def __regularized_spectral_clustering(adj_matrix, tau, n_clusters, algo="scan"):
    """
    :param adj_matrix: adjacency matrix representation of graph where [m][n] >0 if there is edge and [m][n] = weight
    :param n_clusters: cluster partitioning constant
    :param algo: the clustering separation algorithm, possible value kmeans++ or scan
    :return: labels, number of clustering iterations needed, smallest set of cluster found, execution time
    """
    from sklearn.cluster import k_means
    from sklearn.cluster._spectral import discretize

    regularized_laplacian = __regularized_laplacian_matrix(adj_matrix, tau)
    eigen_values, eigen_vectors = __eigen_solver(
        regularized_laplacian, n_clusters=n_clusters
    )
    if algo == "kmeans++":
        _, labels, _, num_iterations = k_means(
            eigen_vectors, n_clusters=n_clusters, return_n_iter=True
        )
    else:
        if n_clusters == 2:  # cluster based on sign
            second_eigen_vector_index = np.argsort(eigen_values)[1]
            second_eigen_vector = eigen_vectors.T[second_eigen_vector_index]
            labels = [
                0 if val <= 0 else 1 for val in second_eigen_vector
            ]  # use only the second eigenvector
            num_iterations = 1
        else:  # bisecting it into k-ways, use all eigenvectors
            labels = discretize(eigen_vectors)
            num_iterations = 20  # assume worst case scenario that it tooks 20 restarts

    smallest_cluster_size = min(np.sum(labels), abs(np.sum(labels) - len(labels)))
    return labels, num_iterations, smallest_cluster_size


def __sklearn_kmeans(adj_matrix, n_clusters):
    from sklearn.cluster import k_means

    _, labels, _, num_iterations = k_means(
        adj_matrix, n_clusters=n_clusters, return_n_iter=True
    )

    smallest_cluster_size = min(np.sum(labels), abs(np.sum(labels) - labels.size))
    return labels, num_iterations, smallest_cluster_size


def __sklearn_spectral_clustering(adj_matrix, n_clusters):
    """
    :param adj_matrix: adjacency matrix representation of graph where [m][n] >0 if there is edge and [m][n] = weight
    :param n_clusters: cluster partitioning constant
    :return: labels, number of clustering iterations needed, smallest set of cluster found, execution time
    """
    from sklearn.cluster import k_means
    from sklearn.neighbors import kneighbors_graph
    from sklearn.manifold import spectral_embedding

    connectivity = kneighbors_graph(adj_matrix, n_neighbors=10, include_self=True)
    affinity_matrix_ = 0.5 * (connectivity + connectivity.T)

    eigen_vectors = spectral_embedding(
        affinity_matrix_,
        n_components=n_clusters,
        eigen_solver="arpack",
        eigen_tol=0.0,
        norm_laplacian=True,
        drop_first=False,
    )

    _, labels, _, num_iterations = k_means(
        eigen_vectors, n_clusters=n_clusters, return_n_iter=True
    )

    smallest_cluster_size = min(np.sum(labels), abs(np.sum(labels) - labels.size))
    return labels, num_iterations, smallest_cluster_size


def rsc_evaluate_graph(
    graph: nx.Graph, n_clusters=2, method="vanilla", percentile=None
):
    """
    Reconsutrction of [1]Understanding Regularized Spectral Clustering via Graph Conductance, Yilin Zhang, Karl Rohe
    :param graph: Graph to be evaluated
    :param n_clusters: How many clusters to look at
    :param method: one among "vanilla", "regularized", "regularized_with_kmeans", "sklearn_spectral_embedding", "sklearn_kmeans", "percentile".
    :param percentile: value in [0, 100]. Mandatory if method="percentile", otherwise None
    :return:
    """
    # Experiment only on undirected graphs
    if graph.is_directed():
        graph = graph.to_undirected()

    # Before computing anything, largest connected component identified and used
    graph = graph.subgraph(max(nx.connected_components(graph), key=len)).copy()

    adj_matrix = nx.to_scipy_sparse_matrix(graph, format="csr")

    if method == "sklearn_spectral_embedding":
        (
            labels,
            num_iterations,
            smallest_cluster_size,
        ) = __sklearn_spectral_clustering(adj_matrix, n_clusters)

    elif method == "sklearn_kmeans":
        labels, num_iterations, smallest_cluster_size = __sklearn_kmeans(
            adj_matrix, n_clusters
        )

    elif method == "vanilla":
        (
            labels,
            num_iterations,
            smallest_cluster_size,
        ) = __regularized_spectral_clustering(adj_matrix, 0, n_clusters)

    elif method == "regularized_with_kmeans":

        graph_degree = graph.degree()
        graph_average_degree = (
            np.sum(val for (node, val) in graph_degree) / graph.number_of_nodes()
        )
        (
            labels,
            num_iterations,
            smallest_cluster_size,
        ) = __regularized_spectral_clustering(
            adj_matrix, graph_average_degree, n_clusters, "kmeans++"
        )
    else:
        graph_degree = graph.degree()
        np.percentile(graph_degree, percentile)
        (
            labels,
            num_iterations,
            smallest_cluster_size,
        ) = __regularized_spectral_clustering(
            adj_matrix, percentile, n_clusters, "scan"
        )

    return labels



import sys
from random import sample
from demon import Demon
from cdlib.algorithms.internal.NodePerception import NodePerception
from cdlib.algorithms.internal import OSSE
import networkx as nx
import numpy as np
from collections import defaultdict
from cdlib import NodeClustering
from cdlib.utils import suppress_stdout, convert_graph_formats, nx_node_integer_mapping
from cdlib.algorithms.internal.CONGO import Congo_
from cdlib.algorithms.internal.CONGA import Conga_
from cdlib.algorithms.internal.LAIS2_nx import LAIS2
from cdlib.algorithms.internal.lfm import LFM_nx
from cdlib.algorithms.internal import LEMON
from cdlib.algorithms.internal.SLPA_nx import slpa_nx
from cdlib.algorithms.internal.multicom import MultiCom
from cdlib.algorithms.internal.PercoMCV import percoMVC
from cdlib.algorithms.internal.LPAM import LPAM
from cdlib.algorithms.internal.core_exp import findCommunities as core_exp_find
from cdlib.algorithms.internal.weightedCommunity import weightedCommunity
from cdlib.algorithms.internal.LPANNI import LPANNI, GraphGenerator
from cdlib.algorithms.internal.DCS import main_dcs
from cdlib.algorithms.internal.UMSTMO import UMSTMO
from cdlib.algorithms.internal.walkscan import WalkSCAN
from cdlib.algorithms.internal.IPCA import i_pca
from cdlib.algorithms.internal.DPCLUS import dp_clus
from cdlib.algorithms.internal.COACH import co_ach
from cdlib.algorithms.internal.graph_entropy import graphentropy
from cdlib.algorithms.internal.EBGC import EBGC
from cdlib.algorithms.internal.EnDNTM import (
    endntm_find_overlap_cluster,
    endntm_evalFuction,
)
import warnings

missing_packages = set()


def __try_load_karate(init=False):
    global karateclub
    if init == True or "karateclub" not in sys.modules:
        try:
            import karateclub

        except ModuleNotFoundError:
            if not init:
                raise ModuleNotFoundError(
                    "Optional dependency not satisfied: install karateclub to use the selected feature."
                )


__try_load_karate(init=True)
if "karateclub" not in sys.modules:
    karateclub = None
    missing_packages.add("karateclub")


try:
    import igraph as ig
except ModuleNotFoundError:
    ig = None
try:
    from angel import Angel
except ModuleNotFoundError:
    Angel = None

try:
    from ASLPAw_package import ASLPAw
except ModuleNotFoundError:
    ASLPAw = None
    missing_packages.add("ASLPAw")

if len(missing_packages) > 0:
    print(
        "Note: to be able to use all overlapping methods, you need to install some additional packages: ",
        missing_packages,
    )

__all__ = [
    "ego_networks",
    "demon",
    "angel",
    "node_perception",
    "overlapping_seed_set_expansion",
    "kclique",
    "lfm",
    "lais2",
    "congo",
    "conga",
    "lemon",
    "slpa",
    "multicom",
    "big_clam",
    "danmf",
    "egonet_splitter",
    "nnsed",
    "mnmf",
    "aslpaw",
    "percomvc",
    "wCommunity",
    "core_expansion",
    "lpanni",
    "lpam",
    "dcs",
    "umstmo",
    "symmnmf",
    "walkscan",
    "endntm",
    "ipca",
    "dpclus",
    "coach",
    "graph_entropy",
    "ebgc",
]


def ego_networks(g_original: object, level: int = 1) -> NodeClustering:
    """
    Ego-networks returns overlapping communities centered at each nodes within a given radius.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param level: extrac communities with all neighbors of distance<=level from a node. Deafault 1
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.ego_networks(G)
    """

    g = convert_graph_formats(g_original, nx.Graph)

    coms = []
    for n in g.nodes():
        coms.append(list(nx.ego_graph(g, n, radius=level).nodes()))
    return NodeClustering(
        coms, g_original, "Ego Network", method_parameters={"level": 1}, overlap=True
    )


def demon(g_original: object, epsilon: float, min_com_size: int = 3) -> NodeClustering:
    """
    Demon is a node-centric bottom-up overlapping community discovery algorithm.
    It leverages ego-network structures and overlapping label propagation to identify micro-scale communities that are subsequently merged in mesoscale ones.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param epsilon: merging threshold in [0,1], default 0.25.
    :param min_com_size: minimum community size, default 3.
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.demon(G, min_com_size=3, epsilon=0.25)

    :References:

    1. Coscia, M., Rossetti, G., Giannotti, F., & Pedreschi, D. (2012, August). `Demon: a local-first discovery method for overlapping communities. <http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.721.1788&rep=rep1&type=pdf/>`_ In Proceedings of the 18th ACM SIGKDD international conference on Knowledge discovery and data mining (pp. 615-623). ACM.

    2. Coscia, M., Rossetti, G., Giannotti, F., & Pedreschi, D. (2014). `Uncovering hierarchical and overlapping communities with a local-first approach. <https://dl.acm.org/citation.cfm?id=2629511/>`_ ACM Transactions on Knowledge Discovery from Data (TKDD), 9(1), 6.

    .. note:: Reference implementation: https://github.com/GiulioRossetti/DEMON

    """

    g = convert_graph_formats(g_original, nx.Graph)

    with suppress_stdout():
        dm = Demon(graph=g, epsilon=epsilon, min_community_size=min_com_size)
        coms = dm.execute()
        coms = [list(c) for c in coms]

    return NodeClustering(
        coms,
        g_original,
        "DEMON",
        method_parameters={"epsilon": epsilon, "min_com_size": min_com_size},
        overlap=True,
    )


def angel(
    g_original: object, threshold: float, min_community_size: int = 3
) -> NodeClustering:
    """
    Angel is a node-centric bottom-up community discovery algorithm.
    It leverages ego-network structures and overlapping label propagation to identify micro-scale communities that are subsequently merged in mesoscale ones.
    Angel is the, faster, successor of Demon.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param threshold: merging threshold in [0,1].
    :param min_community_size: minimum community size, default 3.
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.angel(G, min_com_size=3, threshold=0.25)

    :References:

    1. Rossetti, Giulio. "Exorcising the Demon: Angel, Efficient Node-Centric Community Discovery." International Conference on Complex Networks and Their Applications. Springer, Cham, 2019.

    .. note:: Reference implementation: https://github.com/GiulioRossetti/ANGEL
    """
    if ig is None:
        raise ModuleNotFoundError(
            "Optional dependency not satisfied: install igraph to use the selected feature."
        )
    if Angel is None:
        raise ModuleNotFoundError(
            "Optional dependency not satisfied: install angel-cd library to use the selected feature (likely pip install angel-cd). If using a notebook, you need also to restart your runtime/kernel."
        )

    g = convert_graph_formats(g_original, ig.Graph)
    with suppress_stdout():
        a = Angel(
            graph=g, min_comsize=min_community_size, threshold=threshold, save=False
        )
        coms = a.execute()

    return NodeClustering(
        list(coms.values()),
        g_original,
        "ANGEL",
        method_parameters={
            "threshold": threshold,
            "min_community_size": min_community_size,
        },
        overlap=True,
    )


def node_perception(
    g_original: object,
    threshold: float,
    overlap_threshold: float,
    min_comm_size: int = 3,
) -> NodeClustering:
    """Node perception is based on the idea of joining together small sets of nodes.
    The algorithm first identifies sub-communities corresponding to each node’s perception of the network around it.
    To perform this step, it considers each node individually, and partition that node’s neighbors into communities using some existing community detection method.
    Next, it creates a new network in which every node corresponds to a sub-community, and two nodes are linked if their associated sub-communities overlap by at least some threshold amount.
    Finally, the algorithm identifies overlapping communities in this new network, and for every such community, merge together the associated sub-communities to identify communities in the original network.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param threshold: the tolerance required in order to merge communities
    :param overlap_threshold: the overlap tolerance
    :param min_comm_size: minimum community size default 3
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.node_perception(G, threshold=0.25, overlap_threshold=0.25)

    :References:

    Sucheta Soundarajan and John E. Hopcroft. 2015. `Use of Local Group Information to Identify Communities in Networks. <https://dl.acm.org/citation.cfm?id=2737800.2700404/>`_ ACM Trans. Knowl. Discov. Data 9, 3, Article 21 (April 2015), 27 pages. DOI=http://dx.doi.org/10.1145/2700404

    """
    g = convert_graph_formats(g_original, nx.Graph)
    tp = type(list(g.nodes())[0])

    with suppress_stdout():
        np = NodePerception(
            g,
            sim_threshold=threshold,
            overlap_threshold=overlap_threshold,
            min_comm_size=min_comm_size,
        )
        coms = np.execute()
        if tp != str:
            communities = []
            for c in coms:
                c = list(map(tp, c))
                communities.append(c)
            coms = communities

    return NodeClustering(
        coms,
        g_original,
        "Node Perception",
        method_parameters={
            "threshold": threshold,
            "overlap_threshold": overlap_threshold,
            "min_com_size": min_comm_size,
        },
        overlap=True,
    )


def overlapping_seed_set_expansion(
    g_original: object,
    seeds: list,
    ninf: bool = False,
    expansion: str = "ppr",
    stopping: str = "cond",
    nworkers: int = 1,
    nruns: int = 13,
    alpha: float = 0.99,
    maxexpand: float = float("INF"),
    delta: float = 0.2,
) -> NodeClustering:
    """
    OSSE is an overlapping community detection algorithm optimizing the conductance community score
    The algorithm uses a seed set expansion approach; the key idea is to find good seeds, and then expand these seed sets using the personalized PageRank clustering procedure.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param seeds: Node list
    :param ninf: Neighbourhood Inflation parameter (boolean)
    :param expansion: Seed expansion: ppr or vppr
    :param stopping: Stopping criteria: cond
    :param nworkers: Number of Workers: default 1
    :param nruns: Number of runs: default 13
    :param alpha: alpha value for Personalized PageRank expansion: default 0.99
    :param maxexpand: Maximum expansion allowed for approximate ppr: default INF
    :param delta: Minimum distance parameter for near duplicate communities: default 0.2
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> com = algorithms.overlapping_seed_set_expansion(G)

    :References:

    1.Whang, J. J., Gleich, D. F., & Dhillon, I. S. (2013, October). `Overlapping community detection using seed set expansion. <http://www.cs.utexas.edu/~inderjit/public_papers/overlapping_commumity_cikm13.pdf/>`_ In Proceedings of the 22nd ACM international conference on Conference on information & knowledge management (pp. 2099-2108). ACM.

    .. note:: Reference implementation: https://github.com/pratham16/algorithms-detection-by-seed-expansion
    """

    g = convert_graph_formats(g_original, nx.Graph)

    g, maps = nx_node_integer_mapping(g)
    if maps is not None:
        rev_map = {v: k for k, v in maps.items()}
        seeds = [rev_map[s] for s in seeds]

    if ninf:
        seeds = OSSE.neighbor_inflation(g, seeds)

    communities = OSSE.growclusters(
        g, seeds, expansion, stopping, nworkers, nruns, alpha, maxexpand, False
    )
    communities = OSSE.remove_duplicates(g, communities, delta)
    communities = list(communities)

    if maps is not None:
        coms = []
        for com in communities:
            coms.append([maps[n] for n in com])
        nx.relabel_nodes(g, maps, False)
    else:
        coms = communities

    return NodeClustering(
        coms,
        g_original,
        "Overlapping SSE",
        method_parameters={
            "seeds": seeds,
            "ninf": ninf,
            "expansion": expansion,
            "stopping": stopping,
            "nworkers": nworkers,
            "nruns": nruns,
            "alpha": alpha,
            "maxexpand": maxexpand,
            "delta": delta,
        },
        overlap=True,
    )


def kclique(g_original: object, k: int) -> NodeClustering:
    """
    Find k-clique communities in graph using the percolation method.
    A k-clique community is the union of all cliques of size k that can be reached through adjacent (sharing k-1 nodes) k-cliques.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param k: Size of smallest clique
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> com = algorithms.kclique(G, k=3)

    :References:

    Gergely Palla, Imre Derényi, Illés Farkas1, and Tamás Vicsek, `Uncovering the overlapping community structure of complex networks in nature and society <https://www.nature.com/articles/nature03607/>`_ Nature 435, 814-818, 2005, doi:10.1038/nature03607
    """

    g = convert_graph_formats(g_original, nx.Graph)

    coms = list(nx.algorithms.community.k_clique_communities(g, k))
    coms = [list(x) for x in coms]
    return NodeClustering(
        coms, g_original, "Klique", method_parameters={"k": k}, overlap=True
    )


def lfm(g_original: object, alpha: float) -> NodeClustering:
    """LFM is based on the local optimization of a fitness function.
    It finds both overlapping communities and the hierarchical structure.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param alpha: parameter to controll the size of the communities:  Large values of alpha yield very small communities, small values instead deliver large modules. If alpha is small enough, all nodes end up in the same cluster, the network itself. In most cases, for alpha < 0.5 there is only one community, for alpha > 2 one recovers the smallest communities. A natural choise is alpha =1.
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> com = algorithms.lfm(G, alpha=0.8)

    :References:

    Lancichinetti, Andrea, Santo Fortunato, and János Kertész. `Detecting the overlapping and hierarchical community structure in complex networks <https://arxiv.org/abs/0802.1218/>`_ New Journal of Physics 11.3 (2009): 033015.
    """

    g = convert_graph_formats(g_original, nx.Graph)

    algorithm = LFM_nx(g, alpha)
    coms = algorithm.execute()

    return NodeClustering(
        coms, g_original, "LFM", method_parameters={"alpha": alpha}, overlap=True
    )


def lais2(g_original: object) -> NodeClustering:
    """
    LAIS2 is an overlapping community discovery algorithm based on the density function.
    In the algorithm considers the density of a group is defined as the average density of the communication exchanges between the actors of the group.
    LAIS2 IS composed of two procedures LA (Link Aggregate Algorithm) and IS2 (Iterative Scan Algorithm).


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> com = algorithms.lais2(G)

    :References:

    Baumes, Jeffrey, Mark Goldberg, and Malik Magdon-Ismail. `Efficient identification of overlapping communities. <https://link.springer.com/chapter/10.1007/11427995_3/>`_ International Conference on Intelligence and Security Informatics. Springer, Berlin, Heidelberg, 2005.

    .. note:: Reference implementation: https://github.com/kritishrivastava/CommunityDetection-Project2GDM

    """

    g = convert_graph_formats(g_original, nx.Graph)

    coms = LAIS2(g)
    return NodeClustering(
        coms, g_original, "LAIS2", method_parameters={"": ""}, overlap=True
    )


def congo(
    g_original: object, number_communities: int, height: int = 2
) -> NodeClustering:
    """
    CONGO (CONGA Optimized) is an optimization of the CONGA algortithm.
    The CONGO algorithm is the same as CONGA but using local betweenness. The complete CONGO algorithm is as follows:

    1. Calculate edge betweenness of edges and split betweenness of vertices.
    2. Find edge with maximum edge betweenness or vertex with maximum split betweenness, if greater.
    3. Recalculate edge betweenness and split betweenness:
        a) Subtract betweenness of h-region centred on the removed edge or split vertex.
        b) Remove the edge or split the vertex.
        c) Add betweenness for the same region.
    4. Repeat from step 2 until no edges remain.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param number_communities: the number of communities desired
    :param height: The lengh of the longest shortest paths that CONGO considers, default 2
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> com = algorithms.congo(G, number_communities=3, height=2)

    :References:

    Gregory, Steve. `A fast algorithm to find overlapping communities in networks. <https://link.springer.com/chapter/10.1007/978-3-540-87479-9_45/>`_ Joint European Conference on Machine Learning and Knowledge Discovery in Databases. Springer, Berlin, Heidelberg, 2008.

    .. note:: Reference implementation: https://github.com/Lab41/Circulo/tree/master/circulo/algorithms

    """

    if ig is None:
        raise ModuleNotFoundError(
            "Optional dependency not satisfied: install igraph to use the selected feature."
        )

    g = convert_graph_formats(g_original, ig.Graph)

    communities = Congo_(g, number_communities, height)

    coms = []
    for c in communities:
        coms.append([g.vs[x]["name"] for x in c])

    return NodeClustering(
        coms,
        g_original,
        "Congo",
        method_parameters={"number_communities": number_communities, "height": height},
        overlap=True,
    )


def conga(g_original: object, number_communities: int) -> NodeClustering:
    """
    CONGA (Cluster-Overlap Newman Girvan Algorithm) is an algorithm for discovering overlapping communities.
    It extends the  Girvan and Newman’s algorithm with a specific method of deciding when and how to split vertices. The algorithm is as follows:

    1. Calculate edge betweenness of all edges in network.
    2. Calculate vertex betweenness of vertices, from edge betweennesses.
    3. Find candidate set of vertices: those whose vertex betweenness is greater than the maximum edge betweenness.
    4. If candidate set is non-empty, calculate pair betweennesses of candidate vertices, and then calculate split betweenness of candidate vertices.
    5. Remove edge with maximum edge betweenness or split vertex with maximum split betweenness (if greater).
    6. Recalculate edge betweenness for all remaining edges in same component(s) as removed edge or split vertex.
    7. Repeat from step 2 until no edges remain.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param number_communities: the number of communities desired
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> com = algorithms.conga(G, number_communities=3)

    :References:

    Gregory, Steve. `An algorithm to find overlapping community structure in networks. <https://link.springer.com/chapter/10.1007/978-3-540-74976-9_12/>`_ European Conference on Principles of Data Mining and Knowledge Discovery. Springer, Berlin, Heidelberg, 2007.

    .. note:: Reference implementation: https://github.com/Lab41/Circulo/tree/master/circulo/algorithms
    """

    if ig is None:
        raise ModuleNotFoundError(
            "Optional dependency not satisfied: install igraph to use the selected feature."
        )

    g = convert_graph_formats(g_original, ig.Graph)

    communities = Conga_(g, number_communities=3)
    coms = []
    for c in communities:
        coms.append([g.vs[x]["name"] for x in c])

    return NodeClustering(
        coms,
        g_original,
        "Conga",
        method_parameters={"number_communities": number_communities},
        overlap=True,
    )


def lemon(
    g_original: object,
    seeds: list,
    min_com_size: int = 20,
    max_com_size: int = 50,
    expand_step: int = 6,
    subspace_dim: int = 3,
    walk_steps: int = 3,
    biased: bool = False,
) -> NodeClustering:
    """Lemon is a large scale overlapping community detection method based on local expansion via minimum one norm.

    The algorithm adopts a local expansion method in order to identify the community members from a few exemplary seed members.
    The algorithm finds the community by seeking a sparse vector in the span of the local spectra such that the seeds are in its support. LEMON can achieve the highest detection accuracy among state-of-the-art proposals. The running time depends on the size of the community rather than that of the entire graph.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param seeds: Node list
    :param min_com_size: the minimum size of a single community in the network, default 20
    :param max_com_size: the maximum size of a single community in the network, default 50
    :param expand_step: the step of seed set increasement during expansion process, default 6
    :param subspace_dim: dimension of the subspace; choosing a large dimension is undesirable because it would increase the computation cost of generating local spectra default 3
    :param walk_steps: the number of step for the random walk, default 3
    :param biased: boolean; set if the random walk starting from seed nodes, default False
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> seeds = ["$0$", "$2$", "$3$"]
    >>> coms = algorithms.lemon(G, seeds, min_com_size=2, max_com_size=5)

    :References:

    Yixuan Li, Kun He, David Bindel, John Hopcroft `Uncovering the small community structure in large networks: A local spectral approach. <https://dl.acm.org/citation.cfm?id=2736277.2741676/>`_ Proceedings of the 24th international conference on world wide web. International World Wide Web Conferences Steering Committee, 2015.

    .. note:: Reference implementation: https://github.com/YixuanLi/LEMON
    """

    graph = convert_graph_formats(g_original, nx.Graph)
    graph_m = nx.convert_matrix.to_numpy_array(graph)

    node_to_pos = {n: p for p, n in enumerate(graph.nodes())}
    pos_to_node = {p: n for n, p in node_to_pos.items()}

    seeds = np.array([node_to_pos[s] for s in seeds])

    community = LEMON.lemon(
        graph_m,
        seeds,
        min_com_size,
        max_com_size,
        expand_step,
        subspace_dim=subspace_dim,
        walk_steps=walk_steps,
        biased=biased,
    )

    return NodeClustering(
        [[pos_to_node[n] for n in community]],
        g_original,
        "LEMON",
        method_parameters=dict(
            seeds=str(list(seeds)),
            min_com_size=min_com_size,
            max_com_size=max_com_size,
            expand_step=expand_step,
            subspace_dim=subspace_dim,
            walk_steps=walk_steps,
            biased=biased,
        ),
        overlap=True,
    )


def slpa(g_original: object, t: int = 21, r: float = 0.1) -> NodeClustering:
    """
    SLPA is an overlapping community discovery that extends tha LPA.
    SLPA consists of the following three stages:
    1) the initialization
    2) the evolution
    3) the post-processing


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========


    :param g_original: a networkx/igraph object
    :param t: maximum number of iterations, default 20
    :param r: threshold  ∈ [0, 1]. It is used in the post-processing stage: if the probability of seeing a particular label during the whole process is less than r, this label is deleted from a node’s memory. Default 0.1
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.slpa(G,  t=21, r=0.1)



    :References:

    Xie Jierui, Boleslaw K. Szymanski, and Xiaoming Liu. `Slpa: Uncovering overlapping communities in social networks via a speaker-listener interaction dynamic process. <https://ieeexplore.ieee.org/document/6137400/>`_ Data Mining Workshops (ICDMW), 2011 IEEE 11th International Conference on. IEEE, 2011.

    .. note:: Reference implementation: https://github.com/kbalasu/SLPA
    """

    g = convert_graph_formats(g_original, nx.Graph)

    coms = slpa_nx(g, T=t, r=r)
    return NodeClustering(
        coms, g_original, "SLPA", method_parameters={"T": t, "r": r}, overlap=True
    )


def multicom(g_original: object, seed_node: object) -> NodeClustering:
    """
    MULTICOM is an algorithm for detecting multiple local communities, possibly overlapping, by expanding the initial seed set.
    This algorithm uses local scoring metrics to define an embedding of the graph around the seed set. Based on this embedding, it picks new seeds in the neighborhood of the original seed set, and uses these new seeds to recover multiple communities.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param seed_node: Id of the seed node around which we want to detect communities.
    :return: EdgeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.multicom(G, seed_node=0)

    :References:

    Hollocou, Alexandre, Thomas Bonald, and Marc Lelarge. `Multiple Local Community Detection. <https://hal.archives-ouvertes.fr/hal-01625444/document/>`_ ACM SIGMETRICS Performance Evaluation Review 45.2 (2018): 76-83.

    .. note:: Reference implementation: https://github.com/ahollocou/multicom

    """

    g = convert_graph_formats(g_original, nx.Graph)
    g, maps = nx_node_integer_mapping(g)

    mc = MultiCom(g)
    coms = mc.execute(seed_node)

    if maps is not None:
        communities = []
        for c in coms:
            communities.append([maps[n] for n in c])
        nx.relabel_nodes(g, maps, False)
    else:
        communities = [list(c) for c in coms]

    return NodeClustering(
        communities,
        g_original,
        "Multicom",
        method_parameters={"seeds": seed_node},
        overlap=True,
    )


def big_clam(
    g_original: object,
    dimensions: int = 8,
    iterations: int = 50,
    learning_rate: float = 0.005,
) -> NodeClustering:
    """
    BigClam is an overlapping community detection method that scales to large networks.
    The procedure uses gradient ascent to create an embedding which is used for deciding the node-cluster affiliations.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param dimensions: Number of embedding dimensions. Default 8.
    :param iterations: Number of training iterations. Default 50.
    :param learning_rate: Gradient ascent learning rate. Default is 0.005.
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.big_clam(G)

    :References:

    Yang, Jaewon, and Jure Leskovec. "Overlapping community detection at scale: a nonnegative matrix factorization approach." Proceedings of the sixth ACM international conference on Web search and data mining. 2013.

    .. note:: Reference implementation: https://karateclub.readthedocs.io/
    """
    __try_load_karate()
    g = convert_graph_formats(g_original, nx.Graph)

    model = karateclub.BigClam(
        dimensions=dimensions, iterations=iterations, learning_rate=learning_rate
    )
    model.fit(g)
    members = model.get_memberships()

    # Reshaping the results
    coms_to_node = defaultdict(list)
    for n, c in members.items():
        coms_to_node[c].append(n)

    coms = [list(c) for c in coms_to_node.values()]

    return NodeClustering(
        coms,
        g_original,
        "BigClam",
        method_parameters={
            "dimensions": dimensions,
            "iterations": iterations,
            "learning_rate": learning_rate,
        },
        overlap=True,
    )


def danmf(
    g_original: object,
    layers: tuple = (32, 8),
    pre_iterations: int = 100,
    iterations: int = 100,
    seed: int = 42,
    lamb: float = 0.01,
) -> NodeClustering:
    """
    The procedure uses telescopic non-negative matrix factorization in order to learn a cluster memmbership distribution over nodes. The method can be used in an overlapping and non-overlapping way.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       Yes
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param layers: Autoencoder layer sizes in a list of integers. Default [32, 8].
    :param pre_iterations: Number of pre-training epochs. Default 100.
    :param iterations: Number of training epochs. Default 100.
    :param seed: Random seed for weight initializations. Default 42.
    :param lamb: Regularization parameter. Default 0.01.
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.danmf(G)

    :References:

    Ye, Fanghua, Chuan Chen, and Zibin Zheng. "Deep autoencoder-like nonnegative matrix factorization for community detection." Proceedings of the 27th ACM International Conference on Information and Knowledge Management. 2018.

    .. note:: Reference implementation: https://karateclub.readthedocs.io/
    """

    __try_load_karate()

    g = convert_graph_formats(g_original, nx.Graph)
    model = karateclub.DANMF(layers, pre_iterations, iterations, seed, lamb)

    mapping = {node: i for i, node in enumerate(g.nodes())}
    rev = {i: node for node, i in mapping.items()}
    H = nx.relabel_nodes(g, mapping)

    model.fit(H)
    members = model.get_memberships()

    # Reshaping the results
    coms_to_node = defaultdict(list)
    for n, c in members.items():
        coms_to_node[c].append(rev[n])

    coms = [list(c) for c in coms_to_node.values()]

    return NodeClustering(
        coms,
        g_original,
        "DANMF",
        method_parameters={
            "layers": layers,
            "pre_iteration": pre_iterations,
            "iterations": iterations,
            "seed": seed,
            "lamb": lamb,
        },
        overlap=True,
    )


def egonet_splitter(g_original: object, resolution: float = 1.0) -> NodeClustering:
    """
    The method first creates the egonets of nodes. A persona-graph is created which is clustered by the Louvain method.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param resolution: Resolution parameter of Python Louvain. Default 1.0.
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.egonet_splitter(G)

    :References:

    Epasto, Alessandro, Silvio Lattanzi, and Renato Paes Leme. "Ego-splitting framework: From non-overlapping to overlapping clusters." Proceedings of the 23rd ACM SIGKDD International Conference on Knowledge Discovery and Data Mining. 2017.

    .. note:: Reference implementation: https://karateclub.readthedocs.io/
    """
    __try_load_karate()

    g = convert_graph_formats(g_original, nx.Graph)
    model = karateclub.EgoNetSplitter(resolution=resolution)

    mapping = {node: i for i, node in enumerate(g.nodes())}
    rev = {i: node for node, i in mapping.items()}
    H = nx.relabel_nodes(g, mapping)

    model.fit(H)
    members = model.get_memberships()

    # Reshaping the results
    coms_to_node = defaultdict(list)
    for n, cs in members.items():
        for c in cs:
            coms_to_node[c].append(rev[n])

    coms = [list(c) for c in coms_to_node.values()]

    return NodeClustering(
        coms,
        g_original,
        "EgoNetSplitter",
        method_parameters={"resolution": resolution},
        overlap=True,
    )


def nnsed(
    g_original: object, dimensions: int = 32, iterations: int = 10, seed: int = 42
) -> NodeClustering:
    """
    The procedure uses non-negative matrix factorization in order to learn an unnormalized cluster membership distribution over nodes. The method can be used in an overlapping and non-overlapping way.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param dimensions: Embedding layer size. Default is 32.
    :param iterations: Number of training epochs. Default 10.
    :param seed:  Random seed for weight initializations. Default 42.
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.nnsed(G)

    :References:

    Sun, Bing-Jie, et al. "A non-negative symmetric encoder-decoder approach for community detection." Proceedings of the 2017 ACM on Conference on Information and Knowledge Management. 2017.

    .. note:: Reference implementation: https://karateclub.readthedocs.io/
    """

    __try_load_karate()

    g = convert_graph_formats(g_original, nx.Graph)
    model = karateclub.NNSED(dimensions=dimensions, iterations=iterations, seed=seed)
    model.fit(g)
    members = model.get_memberships()

    # Reshaping the results
    coms_to_node = defaultdict(list)
    for n, c in members.items():
        coms_to_node[c].append(n)

    coms = [list(c) for c in coms_to_node.values()]

    return NodeClustering(
        coms,
        g_original,
        "NNSED",
        method_parameters={
            "dimension": dimensions,
            "iterations": iterations,
            "seed": seed,
        },
        overlap=True,
    )


def mnmf(
    g_original: object,
    dimensions: int = 128,
    clusters: int = 10,
    lambd: float = 0.2,
    alpha: float = 0.05,
    beta: float = 0.05,
    iterations: int = 200,
    lower_control: float = 1e-15,
    eta: float = 5.0,
) -> NodeClustering:
    """
    The procedure uses joint non-negative matrix factorization with modularity based regul;arization in order to learn a cluster memmbership distribution over nodes.
    The method can be used in an overlapping and non-overlapping way.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param dimensions: Number of dimensions. Default is 128.
    :param clusters: Number of clusters. Default is 10.
    :param lambd: KKT penalty. Default is 0.2
    :param alpha: Clustering penalty. Default is 0.05.
    :param beta: Modularity regularization penalty. Default is 0.05.
    :param iterations:  Number of power iterations. Default is 200.
    :param lower_control: Floating point overflow control. Default is 10**-15.
    :param eta: Similarity mixing parameter. Default is 5.0.
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.mnmf(G)

    :References:

    Wang, Xiao, et al. "Community preserving network embedding." Thirty-first AAAI conference on artificial intelligence. 2017.

    .. note:: Reference implementation: https://karateclub.readthedocs.io/
    """
    __try_load_karate()
    g = convert_graph_formats(g_original, nx.Graph)
    model = karateclub.MNMF(
        dimensions=dimensions,
        clusters=clusters,
        lambd=lambd,
        alpha=alpha,
        beta=beta,
        iterations=iterations,
        lower_control=lower_control,
        eta=eta,
    )
    model.fit(g)
    members = model.get_memberships()

    # Reshaping the results
    coms_to_node = defaultdict(list)
    for n, c in members.items():
        coms_to_node[c].append(n)

    coms = [list(c) for c in coms_to_node.values()]

    return NodeClustering(
        coms,
        g_original,
        "MNMF",
        method_parameters={
            "dimension": dimensions,
            "clusters": clusters,
            "lambd": lambd,
            "alpha": alpha,
            "beta": beta,
            "iterations": iterations,
            "lower_control": lower_control,
            "eta": eta,
        },
        overlap=True,
    )


def aslpaw(g_original: object) -> NodeClustering:
    """
    ASLPAw can be used for disjoint and overlapping community detection and works on weighted/unweighted and directed/undirected networks.
    ASLPAw is adaptive with virtually no configuration parameters.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.aslpaw(G)

    :References:

    Xie J, Szymanski B K, Liu X. Slpa: Uncovering Overlapping Communities in Social Networks via a Speaker-Listener Interaction Dynamic Process[C]. IEEE 11th International Conference on Data Mining Workshops (ICDMW). Ancouver, BC: IEEE, 2011: 344–349.

    .. note:: Reference implementation: https://github.com/fsssosei/ASLPAw
    """

    if ASLPAw is None:
        raise ModuleNotFoundError(
            "Optional dependency not satisfied: install gmpy (conda install gmpy2) and ASLPAw (pip install shuffle_graph>=2.1.0 similarity-index-of-label-graph>=2.0.1 ASLPAw>=2.1.0). If using a notebook, you need also to restart your runtime/kernel."
        )

    g = convert_graph_formats(g_original, nx.Graph)
    coms = ASLPAw(g).adj

    communities = defaultdict(list)
    for i, c in coms.items():
        if len(c) > 0:
            for cid in c:
                communities[cid].append(i)

    return NodeClustering(
        list(communities.values()),
        g_original,
        "ASLPAw",
        method_parameters={},
        overlap=True,
    )


def percomvc(g_original: object) -> NodeClustering:
    """
    The PercoMVC approach composes of two steps.
    In the first step, the algorithm attempts to determine all communities that the clique percolation algorithm may find.
    In the second step, the algorithm computes the Eigenvector Centrality method on the output of the first step to measure the influence of network nodes and reduce the rate of the unclassified nodes


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.percomvc(G)

    :References:

    Kasoro, Nathanaël, et al. "PercoMCV: A hybrid approach of community detection in social networks." Procedia Computer Science 151 (2019): 45-52.

    .. note:: Reference implementation: https://github.com/sedjokas/PercoMCV-Code-source
    """
    g = convert_graph_formats(g_original, nx.Graph)
    communities = percoMVC(g)

    return NodeClustering(
        communities, g_original, "PercoMVC", method_parameters={}, overlap=True
    )


def wCommunity(
    g_original: object,
    min_bel_degree: float = 0.7,
    threshold_bel_degree: float = 0.7,
    weightName: str = "weight",
) -> NodeClustering:
    """
    Algorithm to identify overlapping communities in weighted graphs


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       Yes
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param min_bel_degree: the tolerance, in terms of beloging degree, required in order to add a node in a community
    :param threshold_bel_degree: the tolerance, in terms of beloging degree, required in order to add a node in a 'NLU' community
    :param weightName: name of the edge attribute containing the weights
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> nx.set_edge_attributes(G, values=1, name='weight')
    >>> coms = algorithms.wCommunity(G, min_bel_degree=0.6, threshold_bel_degree=0.6)

    :References:

    Chen, D., Shang, M., Lv, Z., & Fu, Y. (2010). Detecting overlapping communities of weighted networks via a local algorithm. Physica A: Statistical Mechanics and its Applications, 389(19), 4177-4187.

    .. note:: Implementation provided by Marco Cardia <cardiamc@gmail.com> and Francesco Sabiu <fsabiu@gmail.com> (Computer Science Dept., University of Pisa, Italy)
    """

    if ig is None:
        raise ModuleNotFoundError(
            "Optional dependency not satisfied: install igraph to use the selected feature."
        )

    g = convert_graph_formats(g_original, ig.Graph)
    # Initialization
    comm = weightedCommunity(
        g,
        min_bel_degree=min_bel_degree,
        threshold_bel_degree=threshold_bel_degree,
        weightName=weightName,
    )
    # Community computation
    comm.computeCommunities()
    # Result
    coms = comm.getCommunities()
    coms = [list(c) for c in coms]

    # renaming and deduplicate communities
    coms_res = set()
    for c in coms:
        coms_res.add(frozenset([g.vs[x]["name"] for x in c]))

    coms_res = [list(c) for c in coms_res]

    return NodeClustering(
        coms_res,
        g_original,
        "wCommunity",
        method_parameters={
            "min_bel_degree": min_bel_degree,
            "threshold_bel_degree": threshold_bel_degree,
            "weightName": weightName,
        },
        overlap=True,
    )


def core_expansion(g_original: object, tolerance: float = 0.0001) -> NodeClustering:
    """
    Core Expansion automatically detect the core of each possible community in the network. Then, it iteratively expand each core by adding the nodes to form the fnal communities. The expansion process is based on the neighborhood overlap measure.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param tolerance: numerical tollerance, default 0.0001
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.core_expansion(G)

    :References:

    Choumane, Ali, Ali Awada, and Ali Harkous. "Core expansion: a new community detection algorithm based on neighborhood overlap." Social Network Analysis and Mining 10 (2020): 1-11.

    .. note:: Reference implementation: https://github.com/pkalkunte18/CoreExpansionAlgorithm
    """
    g = convert_graph_formats(g_original, nx.Graph)
    communities = core_exp_find(g, tolerance)

    return NodeClustering(
        communities,
        g_original,
        "Core Expansion",
        method_parameters={"tolerance": tolerance},
        overlap=True,
    )


def lpanni(g_original: object, threshold: float = 0.1) -> NodeClustering:
    """

    LPANNI (Label Propagation Algorithm with Neighbor Node Influence) detects overlapping community structures by adopting fixed label propagation sequence based on the ascending order of node importance and label update strategy based on neighbor node influence and historical label preferred strategy.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param threshold: Default 0.0001

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.lpanni(G)

    :References:

    Lu, Meilian, et al. "LPANNI: Overlapping community detection using label propagation in large-scale complex networks." IEEE Transactions on Knowledge and Data Engineering 31.9 (2018): 1736-1749.

    .. note:: Reference implementation: https://github.com/wxwmd/LPANNI
    """
    g = convert_graph_formats(g_original, nx.Graph)
    LPANNI(g)
    gen = GraphGenerator(threshold, g)
    communities = [list(c) for c in gen.get_Overlapping_communities()]

    return NodeClustering(
        communities,
        g_original,
        "LPANNI",
        method_parameters={"threshold": threshold},
        overlap=True,
    )


def lpam(
    g_original: object,
    k: int = 2,
    threshold: float = 0.5,
    distance: str = "amp",
    seed: int = 0,
) -> NodeClustering:
    """
    Link Partitioning Around Medoids


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param k: number of clusters
    :param threshold: merging threshold in [0,1], default 0.5
    :param distance: type of distance: "amp" - amplified commute distance, or "cm" - commute distance, or distance matrix between all edges as np ndarray
    :param seed: random seed for k-medoid heuristic
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.lpam(G, k=2, threshold=0.4, distance = "amp")

    :References:

    Alexander Ponomarenko, Leonidas Pitsoulis, Marat Shamshetdinov. "Link Partitioning Around Medoids". https://arxiv.org/abs/1907.08731

    """
    g = convert_graph_formats(g_original, nx.Graph)
    return LPAM(graph=g, k=k, threshold=threshold, distance=distance, seed=seed)


def dcs(g_original: object) -> NodeClustering:
    """
    Divide and Conquer Strategy


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       Yes
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.dcs(G)

    :References:

     Syed Agha Muhammad and Kristof Van Laerhoven. "DCS: Divide and Conquer Strategy For Detecting Overlapping Communities in Social Graphs". https://bit.ly/33m7t3r

    .. note:: Reference implementation: https://github.com/SyedAgha/Divide-and-Conquer/tree/master/DCS_code_and_paper

    """
    g = convert_graph_formats(g_original, nx.Graph)
    communities = main_dcs(g)
    return NodeClustering(
        communities, g_original, "DCS", method_parameters={}, overlap=True
    )


def umstmo(g_original: object) -> NodeClustering:
    """
    Overlapping community detection based on the union of all maximum spanning trees


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :return: NodeClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.umstmo(G)

    :References:

     Asmi, Khawla, Dounia Lotfi, and Mohamed El Marraki. "Overlapping community detection based on the union of all maximum spanning trees." Library Hi Tech (2020).

    .. note:: Reference implementation: https://github.com/khawka/UMSTMO

    """
    g = convert_graph_formats(g_original, nx.Graph)
    communities = UMSTMO(g)
    return NodeClustering(
        communities, g_original, "UMSTMO", method_parameters={}, overlap=True
    )


def symmnmf(
    g_original: object,
    dimensions: int = 32,
    iterations: int = 200,
    rho: float = 100.0,
    seed: int = 42,
) -> NodeClustering:
    """
    The procedure decomposed the second power od the normalized adjacency matrix with an ADMM based non-negative matrix factorization based technique.
    This results in a node embedding and each node is associated with an embedding factor in the created latent space.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param dimensions: Number of dimensions. Default is 32.
    :param iterations:  Number of power iterations. Default is 200.
    :param rho: Regularization tuning parameter. Default is 100.0.
    :param seed: Random seed value. Default is 42.
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.symmnmf(G)

    :References:

    Kuang, Da, Chris Ding, and Haesun Park. "Symmetric nonnegative matrix factorization for graph clustering." Proceedings of the 2012 SIAM international conference on data mining. Society for Industrial and Applied Mathematics, 2012.

    .. note:: Reference implementation: https://karateclub.readthedocs.io/
    """
    __try_load_karate()
    g = convert_graph_formats(g_original, nx.Graph)
    model = karateclub.SymmNMF(
        dimensions=dimensions, iterations=iterations, rho=rho, seed=seed
    )
    model.fit(g)
    members = model.get_memberships()

    # Reshaping the results
    coms_to_node = defaultdict(list)
    for n, c in members.items():
        coms_to_node[c].append(n)

    coms = [list(c) for c in coms_to_node.values()]

    return NodeClustering(
        coms,
        g_original,
        "SymmNMF",
        method_parameters={
            "dimension": dimensions,
            "iterations": iterations,
            "rho": rho,
            "seed": seed,
        },
        overlap=True,
    )


def walkscan(
    g_original: object,
    nb_steps: int = 2,
    eps: float = 0.1,
    min_samples: int = 3,
    init_vector: dict = None,
) -> NodeClustering:
    """
    Random walk community detection method leveraging PageRank node scoring.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param nb_steps: the length of the random walk
    :param eps: DBSCAN eps
    :param min_samples: DBSCAN min_samples
    :param init_vector: dictionary node_id -> initial_probability to initialize the random walk. Default, random selected node with probability set to 1.
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.walkscan(G)

    :References:

    Hollocou, A., Bonald, T., & Lelarge, M. (2016). Improving PageRank for local community detection. arXiv preprint arXiv:1610.08722.

    .. note:: Reference implementation: https://github.com/ahollocou/walkscan
    """
    g = convert_graph_formats(g_original, nx.Graph)
    ws = WalkSCAN(nb_steps=nb_steps, eps=eps, min_samples=min_samples)

    # Initialization vector for the random walk
    if init_vector is None:
        n = sample(list(g.nodes()), 1)[0]
        init_vector = {n: 1}

    ws.detect_communities(g, init_vector)
    coms = [list(c) for c in ws.communities_]

    return NodeClustering(
        coms,
        g_original,
        "walkscan",
        method_parameters={
            "nb_steps": nb_steps,
            "eps": eps,
            "min_samples": min_samples,
            "init_vector": init_vector,
        },
        overlap=True,
    )


def endntm(
    g_original: object, clusterings: list = None, epsilon: float = 2
) -> NodeClustering:
    """
    Overlapping community detection algorithm based on an ensemble  approach with a distributed neighbourhood threshold method (EnDNTM).
    EnDNTM uses pre-partitioned disjoint communities generated by the ensemble mechanism and then analyzes the neighbourhood distribution  of boundary nodes in disjoint communities to detect overlapping communities.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param clusterings: an iterable of Clustering objects (non overlapping node partitions only)
    :param epsilon: neighbourhood threshold, default 2.
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms_l = [algorithms.louvain(G), algorithms.label_propagation(G), algorithms.walktrap(G)]
    >>> coms = algorithms.endntm(G, coms_l)

    :References:

    Jaiswal, R., & Ramanna, S. Detecting overlapping communities using ensemble-based distributed neighbourhood threshold method in social networks. Intelligent Decision Technologies, (2021), doi:10.3233/IDT-200059.

    """
    g = convert_graph_formats(g_original, nx.Graph)
    algClstr = {1: None, 2: []}
    shrDict = {1: 0.0, 2: algClstr}

    if clusterings is None:
        raise ValueError("No precomputed node clusterings provided.")

    for clustering in clusterings:
        if clustering.overlap:
            raise ValueError(
                f"endntm requires non overlapping node clusterings: {clustering.method_name} is overlapping."
            )
        c_name = clustering.method_name
        communities = [set(c) for c in clustering.communities]
        val = endntm_evalFuction(g, communities)

        if val > shrDict[1]:
            shrDict[1] = val
            shrDict[2][1] = c_name
            shrDict[2][2] = communities

    clusters_list = shrDict[2][2]
    overlap_cluster_list = endntm_find_overlap_cluster(g, clusters_list, epsilon)

    coms = [list(c) for c in overlap_cluster_list]

    return NodeClustering(
        coms,
        g_original,
        "endntm",
        method_parameters={
            "clusterings": [clustering.method_name for clustering in clusterings],
            "epsilon": epsilon,
        },
        overlap=True,
    )


def ipca(
    g_original: object,
    weights: str = None,
    t_in: float = 0.5,
) -> NodeClustering:
    """
    IPCA was introduced by Li et al. (2008) as a modiﬁed version of DPClus.
    In contrast to DPClus, this method focuses on the maintaining the diameter of a cluster, deﬁned as the maximum shortest distance between all pairs of vertices, rather than its density.
    In doing so, the seed growth aspect of IPCA emphasizes structural closeness of a predicted protein complex, as well as structural connectivity.

    Like DPClus, IPCA computes local vertex and edge weights by counting the number of common neighbors shared between two vertices.
    However, IPCA calculates these values only once at the beginning of the algorithm, rather than updating them every time a discovered cluster is removed from the graph.
    This allows overlap to occur naturally between clusters, as cluster nodes are not permanently removed from the graph; however, it can also lead to a large amount of cluster overlap.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       Yes
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param weights: label used for the edge weights. Default, None.
    :param t_in:
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.ipca(G)

    :References:

    Li, M., Chen, J., Wang, J., Hu, B., Chen, G. 2008. Modifying the DPClus algorithm for identifying protein complexes based on new topological structures. BMC Bioinformatics 9, 398.


    .. note:: Reference Implementation: https://github.com/trueprice/python-graph-clustering
    """

    g = convert_graph_formats(g_original, nx.Graph)
    clustering = i_pca(
        g,
        weights=weights,
        t_in=t_in,
    )

    return NodeClustering(
        clustering,
        g_original,
        "ipca",
        method_parameters={
            "t_in": t_in,
        },
        overlap=True,
    )


def dpclus(
    g_original: object,
    weights: str = None,
    d_threshold: float = 0.9,
    cp_threshold: float = 0.5,
    overlap: bool = True,
) -> NodeClustering:
    """
    DPClus projects weights onto an unweighted graph using a common neighbors approach.
    In DPClus, the weight of an edge (u, v) is deﬁned as the number of common neighbors between u and v.
    Similarly, the weight of a vertex is its weighted degree – the sum of all edges connected to the vertex-

    DPClus does not natively generate overlapping clusters but does allow for overlapping cluster nodes to be added in a post-processing step.

    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       Yes
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param weights: label used for the edge weights. Default, None.
    :param d_threshold: cluster density threshold, default 0.9
    :param cp_threshold: cluster property threshold, default 0.5
    :param overlap: wheter to output overlapping or crisp communities. Default, True.
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.dpclus(G)

    :References:

    Altaf-Ul-Amin, M., Shinbo, Y., Mihara, K., Kurokawa, K., Kanaya, S. 2006. Development and implementation of an algorithm for detection of protein complexes in large interaction networks. BMC Bioinformatics 7, 207.

    .. note:: Reference Implementation: https://github.com/trueprice/python-graph-clustering
    """

    g = convert_graph_formats(g_original, nx.Graph)
    clustering = dp_clus(
        g,
        weights=weights,
        overlap=overlap,
        d_threshold=d_threshold,
        cp_threshold=cp_threshold,
    )

    return NodeClustering(
        clustering,
        g_original,
        "dpclus",
        method_parameters={"d_threshold": d_threshold, "cp_threshold": cp_threshold},
        overlap=overlap,
    )


def coach(
    g_original: object,
    density_threshold: float = 0.7,
    affinity_threshold: float = 0.225,
    closeness_threshold: float = 0.5,
) -> NodeClustering:
    """
    The motivation behind the core-attachment (CoAch) algorithm  comes from the observation that protein complexes often have a dense core of highly interactive proteins.
    CoAch works in two steps, ﬁrst discovering highly connected regions (“preliminary cores”) of a network and then expanding these regions by adding strongly associated neighbors.

    The algorithm operates with three user-speciﬁed parameters: minimum core density (for preliminary cores), maximum core affinity (similarity threshold for distinct preliminary cores), and minimum neighbor closeness (for attaching non-core neighbors to preliminary cores).


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param density_threshold: minimum core density. Default, 0.7
    :param affinity_threshold: maximum core affinity. Default, 0.225
    :param closeness_threshold:  minimum neighbor closeness. Default, 0.5
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.coach(G)

    :References:

    Wu, M., Li, X., Kwoh, C.-K., Ng, S.-K. A core-attachment based method to detect protein complexes. 2009. In PPI networks. BMC Bioinformatics 10, 169.

    .. note:: Reference Implementation: https://github.com/trueprice/python-graph-clustering
    """

    g = convert_graph_formats(g_original, nx.Graph)
    clustering = co_ach(
        g,
        density_threshold=density_threshold,
        affinity_threshold=affinity_threshold,
        closeness_threshold=closeness_threshold,
    )

    return NodeClustering(
        clustering,
        g_original,
        "coach",
        method_parameters={
            "density_threshold": density_threshold,
            "affinity_threshold": affinity_threshold,
            "closeness_threshold": closeness_threshold,
        },
        overlap=True,
    )


def graph_entropy(
    g_original: object,
    weights: str = None,
) -> NodeClustering:
    """
    This method takes advantage of the use of entropy with regard to information theory.
    Entropy is a measure of uncertainty involved in a random variable.

    This approach uses a new deﬁnition, Graph Entropy, as a measure of structural complexity in a graph.
    This algorithm incorporates a seed-growth technique.
    Unlike the other seed-growth style methods, however, the graph entropy approach does not require any predetermined threshold because it searches for an optimal solution by minimizing graph entropy.

    This method ﬁnds locally optimal clusters with minimal graph entropy.
    A seed vertex is selected at random from a candidate set of seed vertices.
    Then, an initial cluster which is composed of the seed vertex and its immediate neighbors is created.
    Next, the neighbors are iteratively evaluated for removal to minimize the initial entropy of the graph.
    Finally, outer boundary vertices are added recursively if their addition causes the entropy of the graph to decrease.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :param weights: label used for the edge weights.. Default, None
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.graph_entropy(G)

    :References:

    Kenley, E.C., Cho, Y.-R. 2011. Detecting protein complexes and functional modules from protein interaction networks: A graph entropy approach. Proteomics 11, 3835-3844.

    .. note:: Reference Implementation: https://github.com/trueprice/python-graph-clustering
    """

    g = convert_graph_formats(g_original, nx.Graph)
    clustering = graphentropy(g, weight=weights)

    return NodeClustering(
        clustering,
        g_original,
        "graph_entropy",
        method_parameters={},
        overlap=True,
    )


def ebgc(
    g_original: object,
) -> NodeClustering:
    """
    The entropy-based clustering approach finds locally optimal clusters by growing a random seed in a manner that minimizes graph entropy.

    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param g_original: a networkx/igraph object
    :return: NodeClustering object


    :Example:

    >>> from cdlib import algorithms
    >>> import networkx as nx
    >>> G = nx.karate_club_graph()
    >>> coms = algorithms.ebgc(G)

    :References:

    Kenley, Edward Casey, and Young-Rae Cho. "Entropy-based graph clustering: Application to biological and social networks." 2011 IEEE 11th International Conference on Data Mining. IEEE, 2011.

    .. note:: Reference Implementation: https://github.com/SubaiDeng/EBGC-Entropy-Based-Graph-Clustering
    """

    g = convert_graph_formats(g_original, nx.Graph)

    dmap = {n: i for i, n in enumerate(g.nodes)}
    reverse_map = {i: n for n, i in dmap.items()}
    nx.relabel_nodes(g, dmap, False)

    EBGC_cluster = EBGC()
    cluster_result = EBGC_cluster.fit(g)
    _, node_labels = np.nonzero(cluster_result)

    clustering = defaultdict(list)
    for idn, v in enumerate(node_labels):
        clustering[v].append(reverse_map[idn])

    clustering = [c for c in clustering.values()]

    return NodeClustering(
        clustering,
        g_original,
        "ebgc",
        method_parameters={},
        overlap=True,
    )



from cdlib.benchmark.internal import xmark
from cdlib import NodeClustering
from collections import defaultdict

__all__ = ["LFR", "XMark", "GRP", "PP", "RPG", "SBM"]


def LFR(
    n: int,
    tau1: float,
    tau2: float,
    mu: float,
    average_degree: float = None,
    min_degree: int = None,
    max_degree: int = None,
    min_community: int = None,
    max_community: int = None,
    tol: float = 1e-07,
    max_iters: int = 500,
    seed: int = 42,
) -> [object, object]:
    """
    Returns the LFR benchmark graph and planted communities.

    :param n: Number of nodes in the created graph.
    :param tau1: Power law exponent for the degree distribution of the created graph. This value must be strictly greater than one.
    :param tau2: Power law exponent for the community size distribution in the created graph. This value must be strictly greater than one.
    :param mu: Fraction of intra-community edges incident to each node. This value must be in the interval [0, 1].
    :param average_degree: Desired average degree of nodes in the created graph. This value must be in the interval [0, n]. Exactly one of this and min_degree must be specified, otherwise a NetworkXError is raised.
    :param min_degree: Minimum degree of nodes in the created graph. This value must be in the interval [0, n]. Exactly one of this and average_degree must be specified, otherwise a NetworkXError is raised.
    :param max_degree: Maximum degree of nodes in the created graph. If not specified, this is set to n, the total number of nodes in the graph.
    :param min_community: Minimum size of communities in the graph. If not specified, this is set to min_degree.
    :param max_community:  Maximum size of communities in the graph. If not specified, this is set to n, the total number of nodes in the graph.
    :param tol: Tolerance when comparing floats, specifically when comparing average degree values.
    :param max_iters: Maximum number of iterations to try to create the community sizes, degree distribution, and community affiliations.
    :param seed: (integer, random_state, or None (default)) – Indicator of random number generation state. See Randomness.

    :return: A networkx synthetic graph, the set of communities  (NodeClustering object)

    :Example:

    >>> from cdlib.benchmark import LFR
    >>> n = 250
    >>> tau1 = 3
    >>> tau2 = 1.5
    >>> mu = 0.1
    >>> G, coms = LFR(n, tau1, tau2, mu, average_degree=5, min_community=20)

    :References:

    Andrea Lancichinetti, Santo Fortunato, and Filippo Radicchi. “Benchmark graphs for testing community detection algorithms”, Phys. Rev. E 78, 046110 2008

    .. note:: Reference implementation: https://networkx.org/documentation/stable/reference/generated/networkx.generators.community.LFR_benchmark_graph.html#networkx.generators.community.LFR_benchmark_graph
    """

    from networkx.generators.community import LFR_benchmark_graph

    G = LFR_benchmark_graph(
        n=n,
        tau1=tau1,
        tau2=tau2,
        mu=mu,
        average_degree=average_degree,
        min_degree=min_degree,
        max_degree=max_degree,
        min_community=min_community,
        max_community=max_community,
        tol=tol,
        max_iters=max_iters,
        seed=seed,
    )

    communities = {frozenset(G.nodes[v]["community"]) for v in G}
    communities = [list(c) for c in communities]

    coms = NodeClustering(
        communities,
        G,
        "LFR",
        method_parameters={
            "n": n,
            "tau1": tau1,
            "tau2": tau2,
            "mu": mu,
            "average_degree": average_degree,
            "min_degree": min_degree,
            "max_degree": max_degree,
            "min_community": min_community,
            "max_community": max_community,
            "tol": tol,
            "max_iters": max_iters,
            "seed": seed,
        },
    )

    return G, coms


def XMark(
    n: int = 2000,
    gamma: float = 3,
    beta: float = 2,
    m_cat: tuple = ("auto", "auto"),
    theta: float = 0.3,
    mu: float = 0.5,
    avg_k: int = 10,
    min_com: int = 20,
    type_attr: str = "categorical",
) -> [object, object]:
    """
    Returns the XMark benchmark annotated graph and planted communities.

    :param n: Number of nodes in the created graph.
    :param gamma: Power law exponent for the degree distribution of the created graph. This value must be strictly greater than one.
    :param beta: Power law exponent for the community size distribution in the created graph. This value must be strictly greater than one.
    :param m_cat: If the attribute type is categorical, it is the number of values in the domain of the attribute.
    :param m_cont: If the attribute type is continuous, it is the number of peaks in the distribution (at least a bimodal distirbution, i.e., m_cont=2).
    :param theta: If the attribute type is categorical, it specifies the percentage of noise within a cluster.
    :param sigma: If the attribute type is continuous, it is the standard deviation.
    :param mu: Fraction of intra-community edges incident to each node. This value must be in the interval [0, 1].
    :param avg_k: esired average degree of nodes in the created graph. This value must be in the interval [0, n]. Exactly one of this and min_degree must be specified, otherwise a NetworkXError is raised.
    :param min_com: Minimum size of communities in the graph. If not specified, this is set to min_degree.
    :param type_attr: The attribute type. It can be "categorical" or "continuous".

    :return: A networkx synthetic graph, the set of communities  (NodeClustering object)

    :Example:

    >>> from cdlib.benchmark import XMark
    >>> N = 2000
    >>> gamma = 3
    >>> beta = 2
    >>> m_cat = ["auto", "auto"]
    >>> theta = 0.3
    >>> mu = 0.5
    >>> avg_k = 10
    >>> min_com = 20
    >>> g, coms = XMark(n=N, gamma=gamma, beta=beta, mu=mu,
    >>>                           m_cat=m_cat,
    >>>                           theta=theta,
    >>>                           avg_k=avg_k, min_com=min_com,
    >>>                           type_attr="categorical")

    :References:

    Salvatore Citraro, and Giulio Rossetti. “XMark: A Benchmark For Node-Attributed Community Discovery Algorithms”, 2021 (to appear)

    .. note:: Reference implementation: https://github.com/dsalvaz/XMark
    """

    G = xmark.XMark_benchmark(
        n=n,
        tau1=gamma,
        tau2=beta,
        mu=mu,
        labels=m_cat,
        noise=theta,
        average_degree=avg_k,
        min_community=min_com,
        type_attr=type_attr,
    )

    communities = {frozenset(G.nodes[v]["community"]) for v in G}
    communities = [list(c) for c in communities]

    coms = NodeClustering(
        communities,
        G,
        "RDyn",
        method_parameters={
            "n": n,
            "gamma": gamma,
            "beta": beta,
            "mu": mu,
            "m_cat": m_cat,
            "theta": theta,
            "avg_k": avg_k,
            "min_com": min_com,
        },
    )

    return G, coms


def GRP(
    n: int,
    s: float,
    v: float,
    p_in: float,
    p_out: float,
    directed: bool = False,
    seed: object = 42,
) -> [object, object]:
    """
    Generate a Gaussian random partition graph.

    A Gaussian random partition graph is created by creating k partitions
    each with a size drawn from a normal distribution with mean s and variance
    s/v. Nodes are connected within clusters with probability p_in and
    between clusters with probability p_out.

    :param n: Number of nodes in the graph
    :param s: Mean cluster size
    :param v: Shape parameter. The variance of cluster size distribution is s/v.
    :param p_in: Probabilty of intra cluster connection.
    :param p_out: Probability of inter cluster connection.
    :param directed: hether to create a directed graph or not. Boolean, default False
    :param seed: Indicator of random number generation state.

    :return: A networkx synthetic graph, the set of communities  (NodeClustering object)

    :Example:

    >>> from cdlib.benchmark import GRP
    >>> G, coms = GRP(100, 10, 10, 0.25, 0.1)

    :References:

    Ulrik Brandes, Marco Gaertler, Dorothea Wagner, Experiments on Graph Clustering Algorithms,  In the proceedings of the 11th Europ. Symp. Algorithms, 2003.

    .. note:: Reference implementation: https://networkx.org/documentation/stable/reference/generated/networkx.generators.community.gaussian_random_partition_graph.html#networkx.generators.community.gaussian_random_partition_graph
    """

    from networkx.generators.community import gaussian_random_partition_graph

    G = gaussian_random_partition_graph(
        n=n, s=s, v=v, p_in=p_in, p_out=p_out, directed=directed, seed=seed
    )
    communities = defaultdict(list)
    for n, data in G.nodes(data=True):
        communities[data["block"]].append(n)

    coms = NodeClustering(
        list(communities.values()),
        G,
        "GRP",
        method_parameters={
            "n": n,
            "s": s,
            "v": v,
            "p_in": p_in,
            "p_out": p_out,
            "directed": directed,
            "seed": seed,
        },
    )

    return G, coms


def PP(
    l: int, k: int, p_in: float, p_out: float, seed: object = 42, directed: bool = False
) -> [object, object]:
    """
    Returns the planted l-partition graph.

    This model partitions a graph with n=l*k vertices in l groups with k vertices each. Vertices of the same group are linked with a probability p_in, and vertices of different groups are linked with probability p_out.

    :param l: Number of groups
    :param k: Number of vertices in each group
    :param p_in: probability of connecting vertices within a group
    :param p_out:  probability of connected vertices between groups
    :param seed: Indicator of random number generation state.
    :param directed: hether to create a directed graph or not. Boolean, default False

    :return: A networkx synthetic graph, the set of communities  (NodeClustering object)

    :Example:

    >>> from cdlib.benchmark import planted_partitions
    >>> G, coms = planted_partitions(4, 3, 0.5, 0.1, seed=42)

    :References:

    A. Condon, R.M. Karp, Algorithms for graph partitioning on the planted partition model, Random Struct. Algor. 18 (2001) 116-140.
    Santo Fortunato ‘Community Detection in Graphs’ Physical Reports Volume 486, Issue 3-5 p. 75-174. https://arxiv.org/abs/0906.0612

    .. note:: Reference implementation: https://networkx.org/documentation/stable/reference/generated/networkx.generators.community.planted_partition_graph.html#networkx.generators.community.planted_partition_graph
    """
    from networkx.generators.community import planted_partition_graph

    G = planted_partition_graph(
        l=l, k=k, p_in=p_in, p_out=p_out, seed=seed, directed=directed
    )
    communities = defaultdict(list)
    for n, data in G.nodes(data=True):
        communities[data["block"]].append(n)

    coms = NodeClustering(
        list(communities.values()),
        G,
        "planted_partitions",
        method_parameters={
            "l": l,
            "k": k,
            "p_in": p_in,
            "p_out": p_out,
            "seed": seed,
            "directed": directed,
        },
    )

    return G, coms


def RPG(
    sizes: list, p_in: float, p_out: float, seed: object = 42, directed: bool = False
) -> [object, object]:
    """
    Returns the random partition graph with a partition of sizes.

    A partition graph is a graph of communities with sizes defined by s in sizes. Nodes in the same group are connected with probability p_in and nodes of different groups are connected with probability p_out.

    :param sizes: Sizes of groups (list of ints)
    :param p_in: probability of connecting vertices within a group
    :param p_out:  probability of connected vertices between groups
    :param seed: Indicator of random number generation state.
    :param directed: hether to create a directed graph or not. Boolean, default False

    :return: A networkx synthetic graph, the set of communities  (NodeClustering object)

    :Example:

    >>> from cdlib.benchmark import RPG
    >>> G, coms = RPG([10, 10, 10], 0.25, 0.01)

    :References:

    Santo Fortunato ‘Community Detection in Graphs’ Physical Reports Volume 486, Issue 3-5 p. 75-174. https://arxiv.org/abs/0906.0612

    .. note:: Reference implementation: https://networkx.org/documentation/stable/reference/generated/networkx.generators.community.random_partition_graph.html#networkx.generators.community.random_partition_graph
    """

    from networkx.generators.community import random_partition_graph

    G = random_partition_graph(
        sizes=sizes, p_in=p_in, p_out=p_out, seed=seed, directed=directed
    )
    communities = defaultdict(list)
    for n, data in G.nodes(data=True):
        communities[data["block"]].append(n)

    coms = NodeClustering(
        list(communities.values()),
        G,
        "RPG",
        method_parameters={
            "sizes": sizes,
            "p_in": p_in,
            "p_out": p_out,
            "seed": seed,
            "directed": directed,
        },
    )

    return G, coms


def SBM(
    sizes: list,
    p: list,
    nodelist: list = None,
    seed: object = 42,
    directed: bool = False,
    selfloops: bool = False,
    sparse: bool = True,
) -> [object, object]:
    """
    Returns a stochastic block model graph.

    This model partitions the nodes in blocks of arbitrary sizes, and places edges between pairs of nodes independently, with a probability that depends on the blocks.

    :param sizes: Sizes of blocks (list of ints)
    :param p: Element (r,s) gives the density of edges going from the nodes of group r to nodes of group s. p must match the number of groups (len(sizes) == len(p)), and it must be symmetric if the graph is undirected. (List of floats)
    :param nodelist: The block tags are assigned according to the node identifiers in nodelist. If nodelist is None, then the ordering is the range [0,sum(sizes)-1]. Optional, default None.
    :param seed: Indicator of random number generation state.
    :param directed: hether to create a directed graph or not. Boolean, default False.
    :param selfloops: Whether to include self-loops or not. Optional, default False.
    :param sparse: Use the sparse heuristic to speed up the generator. Optional, default True.

    :return: A networkx synthetic graph, the set of communities  (NodeClustering object)

    :Example:

    >>> from cdlib.benchmark import SBM
    >>> sizes = [75, 75, 300]
    >>> probs = [[0.25, 0.05, 0.02], [0.05, 0.35, 0.07], [0.02, 0.07, 0.40]]
    >>> G, coms = SBM(sizes, probs, seed=0)

    :References:

    Holland, P. W., Laskey, K. B., & Leinhardt, S., “Stochastic blockmodels: First steps”, Social networks, 5(2), 109-137, 1983.

    .. note:: Reference implementation: https://networkx.org/documentation/stable/reference/generated/networkx.generators.community.stochastic_block_model.html#networkx.generators.community.stochastic_block_model
    """

    from networkx.generators.community import stochastic_block_model

    G = stochastic_block_model(
        sizes=sizes,
        p=p,
        nodelist=nodelist,
        seed=seed,
        directed=directed,
        selfloops=selfloops,
        sparse=sparse,
    )
    communities = defaultdict(list)
    for n, data in G.nodes(data=True):
        communities[data["block"]].append(n)

    coms = NodeClustering(
        list(communities.values()),
        G,
        "SBM",
        method_parameters={
            "sizes": sizes,
            "p": p,
            "nodelist": nodelist,
            "seed": seed,
            "directed": directed,
            "selfloops": selfloops,
            "sparse": sparse,
        },
    )

    return G, coms



from cdlib.classes.node_clustering import NodeClustering


class BiNodeClustering(NodeClustering):

    """Bipartite Node Communities representation.

    :param left_communities: list of left communities
    :param right_communities: list of right communities
    :param graph: a networkx/igraph object
    :param method_name: community discovery algorithm name
    :param method_parameters: configuration for the community discovery algorithm used
    :param overlap: boolean, whether the partition is overlapping or not
    """

    def __init__(
        self,
        left_communities: list,
        right_communities: list,
        graph: object,
        method_name: str = "",
        method_parameters: dict = None,
        overlap: bool = False,
    ):
        self.left_communities = left_communities
        self.right_communities = right_communities
        super().__init__(
            self.left_communities + self.right_communities,
            graph,
            method_name,
            method_parameters,
            overlap,
        )



import math
import random
import sys
import operator
from collections import defaultdict


def principled(g, cluster_count):

    clusters = range(cluster_count)
    nodes = list(g.nodes())
    edges = list(g.edges())

    # Do the actual clustering
    k = defaultdict(lambda: defaultdict(lambda: random.random()))
    for n in range(1, 100):
        kappa = defaultdict(lambda: 0)
        k_prime = defaultdict(lambda: defaultdict(lambda: 0))
        for z in clusters:
            for i in nodes:
                kappa[z] += k[i][z]

        for i, j in edges:
            D = 0
            for z in clusters:
                D += (k[i][z] * k[j][z]) / kappa[z]

            for z in clusters:
                q = (k[i][z] * k[j][z]) / (D * kappa[z])
                k_prime[i][z] += q
                k_prime[j][z] += q
        k = k_prime

    # Output per-node community membership
    communities = defaultdict(list)
    allocation_matrix = {}
    for i in nodes:
        expected_colors = k[i].items()
        expected_degree = sum([c[1] for c in expected_colors])
        membership = map(
            lambda x: (x[0], 100 * x[1] / expected_degree), expected_colors
        )
        allocation_matrix[i] = {cluster: weight for cluster, weight in membership}
        communities[
            max(allocation_matrix[i].items(), key=operator.itemgetter(1))[0]
        ].append(i)

    return list(communities.values()), allocation_matrix



"""
Name :: UnityId
Shruti Kuber :: skuber
Abhishek Lingwal :: aslingwa
Raunaq Saxena :: rsaxena

Chen, Jie, and Yousef Saad.
"Dense subgraph extraction with application to algorithms detection."
IEEE Transactions on Knowledge and Data Engineering 24.7 (2012): 1216-1230.

Reference internal: https://github.com/imabhishekl/CSC591_Community_Detection
"""

import numpy as np
import networkx as nx
import scipy


class __Node:
    def __init__(self, name):
        self.name = name
        self.left = None
        self.right = None
        self.parent = None
        self.num_edges = 0
        self.vertices = set()
        self.density = 0


class __Tree:
    def __init__(self):
        self.root = None

    def findLCA_Node(self, src_node, dest_node):
        while src_node is not None:
            if dest_node.name in src_node.vertices:
                return src_node
            src_node = src_node.parent
        return None

    def count_vertices_and_edges(self, edges_list, nodes_list):
        for edge in edges_list:
            lca_node = None

            src_node = nodes_list[edge[0]] if nodes_list.__contains__(edge[0]) else None
            dst_node = nodes_list[edge[1]] if nodes_list.__contains__(edge[1]) else None
            if src_node is not None and dst_node is not None:
                lca_node = self.findLCA_Node(src_node, dst_node)

            if lca_node is not None:
                lca_node.num_edges = lca_node.num_edges + 1

    def count_vertices_and_edges_wrap(self, root):
        if root.left is not None and root.right is not None:
            self.count_vertices_and_edges_wrap(root.left)
            self.count_vertices_and_edges_wrap(root.right)
        if root.left is not None and root.right is not None:
            root.num_edges = root.left.num_edges + root.right.num_edges + root.num_edges
        # print root.name, root.num_edges

    def compute_density(self, root):
        if root.left is None and root.right is None:
            return
        total_vertices = float(len(root.vertices))
        max_vertices = total_vertices * (total_vertices - 1) / 2
        root.density = root.num_edges / max_vertices
        self.compute_density(root.left)
        self.compute_density(root.right)

    def extract_sub_graph(self, root, min_density, result):

        if root is None:
            return
        if root.density > min_density:
            com = []
            for elem in list(root.vertices):
                com.append(elem)
            result.append(com)
        else:
            self.extract_sub_graph(root.left, min_density, result)
            self.extract_sub_graph(root.right, min_density, result)


def __make_set(r):
    r.parent = None
    r.vertices.add(r.name)


def __set_find(r):
    while r.parent != None:
        r = r.parent
    return r


# Building a new Node as Union of two sets
def __set_union(x, y):
    r = __Node("P" + str(x.name) + str(y.name))
    r.left = x
    r.right = y
    x.parent = r
    y.parent = r
    r.vertices = r.vertices.union(x.vertices, y.vertices)
    return r


def GDMP2(graph, min_threshold=0.75):

    A = nx.adjacency_matrix(graph)
    adj_matrix = A.todense()

    M = np.zeros(adj_matrix.shape)

    row, col = adj_matrix.shape

    # Building similarity function matrix, ie, Cosine Function matrix of all Column Vectors

    for x in range(0, row):
        for y in range(x, col):
            M[x][y] = 1 - scipy.spatial.distance.cosine(
                adj_matrix[:, x], adj_matrix[:, y]
            )

    tuples = []
    # On basis of zero graph
    min_value = 1 if min(graph.nodes()) > 0 else 0

    # Considering only non zero values
    for (x, y), value in np.ndenumerate(M):
        if value != 0 and x != y:
            tuples.append(((x + min_value, y + min_value), value))

    C = sorted(tuples, key=lambda x: x[1])
    t = np.count_nonzero(adj_matrix)
    C = C[-t:]
    ln = len(C)
    ln = ln - 1

    nodes = dict()
    root_nodes = set()
    tree = __Tree()

    for index in range(ln, -1, -1):
        vertices, value = C[index]
        i, j = vertices
        if nodes.__contains__(i) is False:
            a = __Node(i)
            __make_set(a)
            nodes[i] = a
        if nodes.__contains__(j) is False:
            a = __Node(j)
            __make_set(a)
            nodes[j] = a

        i = nodes[i]
        j = nodes[j]
        ri = __set_find(i)
        rj = __set_find(j)
        if ri.vertices != rj.vertices:
            temp_root = __set_union(ri, rj)
            root_nodes.add(temp_root)

    root_nodes = filter(lambda entry: entry.parent is None, list(root_nodes))

    for temp_roots in root_nodes:
        tree.root = temp_roots

        # Counting number of vertices and Edges
        tree.count_vertices_and_edges(graph.edges(), nodes)

        # Summing up number of edges of children to parent
        tree.count_vertices_and_edges_wrap(tree.root)

        # Computing density of Tree Nodes
        tree.compute_density(tree.root)

        # Filtering Nodes as Per Density Threshold
        communities = []
        tree.extract_sub_graph(tree.root, min_threshold, communities)
        return communities



from cdlib.classes.clustering import Clustering


class NamedClustering(Clustering):
    """Node Communities representation.

    :param communities: dict of named communities <community_id, node_list>
    :param graph: a networkx/igraph object
    :param method_name: community discovery algorithm name
    :param method_parameters: configuration for the community discovery algorithm used
    :param overlap: boolean, whether the partition is overlapping or not
    """

    def __init__(
        self,
        communities: list,
        graph: object,
        method_name: str,
        method_parameters: dict = None,
        overlap: bool = False,
    ):
        self.named_communities = communities
        flat_coms = list(communities.values())
        super().__init__(flat_coms, graph, method_name, method_parameters, overlap)



"""
Baumes, Jeffrey, Mark Goldberg, and Malik Magdon-Ismail.
"Efficient identification of overlapping communities."
International Conference on Intelligence and Security Informatics.
Springer, Berlin, Heidelberg, 2005.

Reference internal
https://github.com/kritishrivastava/CommunityDetection-Project2GDM

"""

import networkx as nx


def __weight(community):
    """

    :param community: a subgraph/algorithms in the graph
    :return: weight of the algorithms (using the formula mentioned in the paper)
    """
    if nx.number_of_nodes(community) == 0:
        return 0
    else:
        return float(2 * nx.number_of_edges(community) / nx.number_of_nodes(community))


def __order_nodes(graph):
    """

    :param graph:
    :return: list of nodes sorted in the decreasing order of their page rank
    """
    dict_of_nodes = nx.pagerank(graph)
    ordered_nodes = dict_of_nodes.items()
    ordered_nodes = sorted(ordered_nodes, reverse=True, key=__get_key)
    return ordered_nodes


def __get_key(node):
    """

    :param node: list containing node name and its page rank
    :return: return rank of the node
    """
    return node[1]


def __list_aggregate(graph):
    """

    :param graph:
    :return: a group of clusters (initial guesses to be fed into the second algorithm)
    order the vertices using page rank
    """
    ordered_nodes = __order_nodes(graph)
    coms = []
    for i in ordered_nodes:
        added = False

        for c in coms:
            # Add the node and see if the weight of the cluster increases
            temp1 = graph.subgraph(c)
            cc = list(c)
            cc.append(i[0])
            temp2 = graph.subgraph(cc)

            # If weight increases, add the node to the cluster
            if __weight(temp2) > __weight(temp1):
                added = True
                c.append(i[0])
        if not added:
            coms.append([i[0]])
    return coms


def __is2(cluster, graph):
    """

    :param cluster:cluster to be improved
    :param graph:
    :return: improved cluster
    """
    c = graph.subgraph(cluster)
    initial_weight = __weight(c)
    increased = True
    while increased:
        list_of_nodes = cluster

        for vertex in c.nodes():
            # Get adjacent nodes
            adjacent_nodes = graph.neighbors(vertex)
            # Adding all adjacent nodes to the cluster
            list_of_nodes = list(set(list_of_nodes).union(set(adjacent_nodes)))

        for vertex in list_of_nodes:
            list_of_nodes = list(c.nodes())

            # If the vertex was a part of inital cluster
            if vertex in c.nodes():
                # Remove vertex from the cluster
                list_of_nodes.remove(vertex)

            # If the vertex is one of the recently added neighbours
            else:
                # Add vertex to the cluster
                list_of_nodes.append(vertex)
            c_dash = graph.subgraph(list_of_nodes)

            if __weight(c_dash) > __weight(c):
                c = c_dash.copy()

        new_weight = __weight(c)

        if new_weight == initial_weight:
            increased = False
        else:
            initial_weight = new_weight
    return c


def LAIS2(graph):
    """

    :param graph:
    :return:
    """

    # Generate initial "guesses" for clusters using Link Aggregate algorithm
    initial_clusters = __list_aggregate(graph)

    # Get final clusters using Improved Iterative Scan Algorithm
    final_clusters = []
    initial_clusters_without_duplicates = []
    for cluster in initial_clusters:
        cluster = sorted(cluster)
        if cluster not in initial_clusters_without_duplicates:
            initial_clusters_without_duplicates.append(cluster)
            updated_cluster = __is2(cluster, graph)
            final_clusters.append(updated_cluster.nodes())

    # Removing duplicate clusters and printing output to a file
    final_clusters_without_duplicates = []
    for cluster in final_clusters:
        cluster = sorted(cluster)
        if cluster not in final_clusters_without_duplicates:
            final_clusters_without_duplicates.append(cluster)

    return final_clusters_without_duplicates



"""
Implementation of the bigClAM algorithm.
Throughout the code, we will use tho following variables
  * F refers to the membership preference matrix. It's in [NUM_PERSONS, NUM_COMMUNITIES]
   so index (p,c) indicates the preference of person p for algorithms c.
  * A refers to the adjency matrix, also named friend matrix or edge set. It's in [NUM_PERSONS, NUM_PERSONS]
    so index (i,j) indicates is 1 when person i and person j are friends.
"""

import numpy as np
import networkx as nx


def sigm(x):
    return np.divide(np.exp(-1.0 * x), 1.0 - np.exp(-1.0 * x))


def log_likelihood(F, A):
    """implements equation 2 of
    https://cs.stanford.edu/people/jure/pubs/bigclam-wsdm13.pdf"""
    A_soft = F.dot(F.T)

    # Next two lines are multiplied with the adjacency matrix, A
    # A is a {0,1} matrix, so we zero out all elements not contributing to the sum
    FIRST_PART = A * np.log(1.0 - np.exp(-1.0 * A_soft))
    sum_edges = np.sum(FIRST_PART)
    SECOND_PART = (1 - A) * A_soft
    sum_nedges = np.sum(SECOND_PART)

    log_likeli = sum_edges - sum_nedges
    return log_likeli


def gradient(F, A, i):
    """Implements equation 3 of
    https://cs.stanford.edu/people/jure/pubs/bigclam-wsdm13.pdf

      * i indicates the row under consideration

    The many forloops in this function can be optimized, but for
    educational purposes we write them out clearly
    """
    N, C = F.shape
    neighbours = np.where(A[i])
    nneighbours = np.where(1 - A[i])

    sum_neigh = np.zeros((C,))
    for nb in neighbours[0]:
        dotproduct = F[nb].dot(F[i])
        sum_neigh += F[nb] * sigm(dotproduct)

    sum_nneigh = np.zeros((C,))
    # Speed up this computation using eq.4
    for nnb in nneighbours[0]:
        sum_nneigh += F[nnb]

    grad = sum_neigh - sum_nneigh
    return grad


def train(A, C, iterations=100):
    # initialize an F
    N = A.shape[0]
    F = np.random.rand(N, C)

    for n in range(iterations):
        for person in range(N):
            grad = gradient(F, A, person)

            F[person] += 0.005 * grad

            F[person] = np.maximum(0.001, F[person])  # F should be nonnegative
        log_likelihood(F, A)
    return F


def big_Clam(graph, number_communities):
    adj = nx.to_numpy_matrix(graph)
    F = train(adj, number_communities)
    F_argmax = np.argmax(F, 1)
    dict_communities = {}
    for i in range(0, number_communities):
        dict_communities[i] = []
    for node, com in zip(graph.nodes(), F_argmax):
        dict_communities[com].append(node)

    list_communities = []
    for com in dict_communities:
        list_communities.append(dict_communities[com])

    return list_communities



import networkx as nx
import numpy as np
from functools import reduce


class Leader_Identification(object):
    def __init__(self, leader_epsilon):
        self.overlap = leader_epsilon

    @staticmethod
    def __inclusion(c1, c2):
        """
        :param c1: node neighbors
        :param c2: node neighbors
        """

        intersection = set(c2) & set(c1)
        smaller_set = min(len(c1), len(c2))
        return len(intersection) / float(smaller_set)

    def __search_leader(self, G, leaders, is_leader, epsilon):
        """
        :param G: G: the networkx graph on which perform DCS
        :param leadrs:list of leaders
        :param is_leader: check if this node satisfies our condition
        :param dlt_nodes: nodes that failed to satisfy our condition in previous iteration
        :param epsilon: the tolerance required in order to merge communities
        """

        com_val = []
        find_edge = []

        node_nbrs = list(G.neighbors(is_leader))

        if len(node_nbrs) > 2:

            for _, leader in enumerate(leaders):
                lead_nbrs = list(G.neighbors(leader))
                find_edge = np.append(find_edge, G.has_edge(is_leader, leader))
                com_val = np.append(com_val, self.__inclusion(node_nbrs, lead_nbrs))

        if len(com_val) > 0:
            return (
                len(com_val),
                len(np.where(np.array(com_val) <= epsilon)[0]),
                len(np.where(find_edge > 0)[0]),
            )
        else:
            if len(find_edge) > 0:
                return len(com_val), 0, len(np.where(find_edge > 0)[0])
            else:
                return len(com_val), 0, 0

    def leader_finding(self, G):
        """
        :param G: G: the networkx graph on which perform DCS
        """

        nod = [
            nx.degree(G, i) + sum(dict(nx.degree(G, G.neighbors(i))).values())
            for i in G.nodes()
        ]
        sort_with_extended_degree = sorted(
            np.column_stack((nod, G.nodes())), key=lambda x: x[0], reverse=True
        )
        leaders = [sort_with_extended_degree[i][1] for i in range(2)]

        for j, k in enumerate(sort_with_extended_degree[2:-1]):

            leader_condition = self.__search_leader(G, leaders, k[1], self.overlap)

            if (
                leader_condition[1] > 0
                and ((leader_condition[0] - leader_condition[1]) <= 2)
                and leader_condition[2] <= 1
            ):
                leaders.append(k[1])

        return leaders


class Merging(object):
    def __init__(self):
        """
        Constructor
        """

    @staticmethod
    def __generalized_inclusion(c1, c2, epsilon):
        """
        :param c1: community
        :param c2: community
        """

        res = None
        intersection = set(c2) & set(c1)
        smaller_set = min(len(c1), len(c2))

        if len(intersection) == 0:
            return None

        if not smaller_set == 0:
            res = float(len(intersection)) / float(smaller_set)

        if res >= epsilon:
            union = set(c2) | set(c1)
            return union

    def merge_communities(self, communities, actual_community, epsilon):

        """
        :param communities: dictionary of communities
        :param actual_community: a community
        :param epsilon: the tolerance required in order to merge communities
        """

        if tuple(actual_community) in communities:
            return communities

        else:

            inserted = False

            for test_community in communities.items():

                union = self.__generalized_inclusion(
                    actual_community, test_community[0], epsilon
                )

                if union is not None:
                    communities.pop(test_community[0])
                    communities[tuple(sorted(union))] = 0
                    inserted = True
                    break

            if not inserted:
                communities[tuple(sorted(actual_community))] = 0

        return communities


class Community(Merging):
    def __init__(self):
        """
        Constructor
        """

        Merging.__init__(self)
        self.all_communities = {}

    def execute(self, G, leader_nodes, epsilon, depth):
        """
        Execute CD algorithm
        :param leader_nodes:
        :param G: the networkx graph on which perform detection
        :param epsilon: the tolerance required in order to merge communities (default 0.5)
        :param depth: the level of depth for neighborhood extraction (1 or 2)
        """

        self.G = G
        self.epsilon = epsilon
        self.depth = depth
        self._communities(leader_nodes, G, nx.density(G), self.depth)

        return self.all_communities

    def _communities(self, leaders, G, graph_density, depth):
        """
        :param leader: leading nodes
        :param G: graph
        :param graph_density: density of the whole graph
        :param depth: neighboorhood order
        """

        for _, i in enumerate(leaders):

            order_neigborhod = 1
            new = []
            community_list = []
            dlt_nodes = []
            ego_nodes = set()

            while order_neigborhod <= depth:

                ego_nodes = self.__without_ego(
                    self.G, ego_nodes, order_neigborhod, dlt_nodes, i
                )
                extract_subgraph = G.subgraph(ego_nodes)
                edges_in_subgraph = extract_subgraph.size()
                nodes_to_be_checked_for_community = list(
                    set(ego_nodes).difference({i}, set(new))
                )
                new.extend(nodes_to_be_checked_for_community)
                outside_cc_edges = (
                    sum(dict(nx.degree(G, ego_nodes)).values()) - 2 * edges_in_subgraph
                )
                with_node_conductance = outside_cc_edges / (
                    (2.0 * edges_in_subgraph) + outside_cc_edges
                )

                for _, nods in enumerate(nodes_to_be_checked_for_community):

                    in_comm = extract_subgraph.degree(nods)
                    out_comm = G.degree(nods) - in_comm
                    outside_edges_of_a_node = edges_in_subgraph - in_comm
                    without_node_cf = (outside_cc_edges - out_comm) / (
                        0.001
                        + (2.0 * outside_edges_of_a_node)
                        + (outside_cc_edges - out_comm)
                    )
                    conductance_score_node = without_node_cf - with_node_conductance
                    subgraph_density_without_node = self.__generalized(
                        outside_edges_of_a_node, len(extract_subgraph) - 1
                    )
                    density_score_node = (
                        nx.density(extract_subgraph)
                        - subgraph_density_without_node
                        + graph_density
                    )

                    """condition for a node to be included in a community"""
                    if conductance_score_node >= 0 and density_score_node >= 0:
                        community_list.append(nods)
                    else:
                        dlt_nodes.append(nods)

                order_neigborhod += 1

            community_list.append(i)

            if len(community_list) <= 6:
                community_list.extend(extract_subgraph.nodes())

            self.all_communities = Merging.merge_communities(
                self, self.all_communities, list(set(community_list)), self.epsilon
            )
        return

    @staticmethod
    def __generalized(d1, d2):
        """
        :param d1: edges
        :param d2: nodes
        """

        if d2 == 1:
            return 0

        return (2.0 * d1) / (d2 * (d2 - 1))

    @staticmethod
    def __without_ego(G, list_nodes, root, dlt_nodes, leader):
        """
        :param G: graph
        :list_nodes: nodes included in the previous iteration
        :root: depth level
        :dlt_nodes: nodes that failed to satisfy our condition in previous iteration
        :leader: root node
        """

        if root == 1:
            nodes = {leader}
            nodes.update(G.neighbors(leader))
            return nodes
        else:
            remove_unimportant_nodes = list(set(list_nodes).difference(set(dlt_nodes)))
            nodes = set()
            nodes.update(remove_unimportant_nodes)
            for n in remove_unimportant_nodes:
                nodes.update(nx.neighbors(G, n))
            return set(nodes).difference(set(dlt_nodes))


class Bridge(Merging):
    def __init__(self, size=None):
        """
        Constructor
        """

        Merging.__init__(self)
        self.size = size

    @staticmethod
    def __assign_to_biggest_module(comm, community_list):
        comm = list(comm)
        comm.pop(np.argmax(community_list))
        community_list.remove(max(community_list))
        return list(reduce(lambda x, y: x + y, comm))

    @staticmethod
    def __delete_edges(graph, nolist, node):

        select_edges = set.intersection(set(nolist), set(nx.neighbors(graph, node)))
        select_edges_to_remove = [
            (j, node)
            for j in select_edges
            if (nx.degree(graph, node) > 1 and nx.degree(graph, j) > 1)
        ]

        graph.remove_edges_from(select_edges_to_remove)
        return graph

    def create_ap_points(self, G):
        """
        this method creates weak articulation points
        :param G: the networkx graph on which perform detection
        """

        for j in G:

            egonet = G.subgraph(nx.all_neighbors(G, j))
            local = nx.connected_components(egonet)
            extract_components = list(local)

            # to avoid whiskers, we verify this condition and it produced good results for >5
            if len(egonet.nodes()) > 5:

                all_communities = {}

                for k in extract_components:
                    for l in k:
                        # small value of epsilon or merging parameter to identiy divserse
                        # second-order neighborhood in the graph
                        all_communities = Merging.merge_communities(
                            self, all_communities, nx.neighbors(G, l), 0.30
                        )

                communities_counter = [len(x) for x in all_communities.keys()]

                if sum(np.bincount(communities_counter)[3:]) > 1:
                    # this process is similar for this function and bridge_function
                    tobedel = self.__assign_to_biggest_module(
                        all_communities.keys(), communities_counter
                    )
                    G = nx.Graph(G)
                    self.__delete_edges(G, tobedel, j)

        return G

    def bridge_function(self, G):
        """
        bridge function to identify weak nodes of the graph
        :param G: the networkx graph on which perform detection
        """

        components = nx.connected_components(G)
        big_mod = {}
        normal_mod = {}
        bigmodcounter = normmodcounter = 0

        for _, i in enumerate(components):
            i = nx.subgraph(G, i)
            for _, j in enumerate(i):

                egonet = i.subgraph(nx.all_neighbors(i, j))
                local = nx.connected_components(egonet)
                conn_comp = list(local)

                if len(conn_comp) > 1:

                    all_communities = {}

                    for k in conn_comp:

                        egos_local_communities = set(k)
                        for l in k:
                            egos_local_communities.update(nx.neighbors(i, l))

                        egos_local_communities.remove(j)

                        # merge two list if they have even a single member in common
                        all_communities = Merging.merge_communities(
                            self, all_communities, list(egos_local_communities), 0
                        )

                    communities_counter = [len(x) for x in all_communities.keys()]

                    if sum(np.bincount(communities_counter)[3:]) > 1:
                        # if big modules identified then remove the edges and assign
                        # the node to the largest component connected to it
                        tobedel = self.__assign_to_biggest_module(
                            all_communities.keys(), communities_counter
                        )
                        i = nx.Graph(i)
                        self.__delete_edges(i, tobedel, j)

            assign_modules = list(nx.connected_components(i))

            # check modules size and assign them
            for i, mod in enumerate(assign_modules):

                if len(mod) > self.size:
                    big_mod[bigmodcounter] = mod
                    bigmodcounter += 1

                else:
                    normal_mod[normmodcounter] = mod
                    normmodcounter += 1

        return big_mod, normal_mod


def main_dcs(G):
    # local centrality to identify meaningful modules of the graph
    # we set 3000 as a module size as discusssed in our paper
    local_centrality = Bridge(size=3000)
    bigmod, normmod = local_centrality.bridge_function(G)

    for _, j in enumerate(bigmod.values()):

        G1 = G.subgraph(j)
        G2 = local_centrality.create_ap_points(G1)
        _, norm_modules = local_centrality.bridge_function(G2)

        for i, k in enumerate(norm_modules.values(), start=len(normmod)):
            normmod[i] = k

    # leader identification and community spreading phase
    communities = {}
    comm_counter = 0

    for j, i in enumerate(normmod.values()):

        # for very small modules, we search for its neighbors and merge them in the communities
        if len(i) <= 9:

            sma = set(i)
            for k in i:
                sma.update(nx.neighbors(G, k))

            communities[comm_counter] = sma
            comm_counter += 1

        else:

            # to avoid local heterogeneity we set epsilon (merging value)
            # of 0.75 for big modules and 0.30 for small modules
            if 10 <= len(i) <= 500:
                heterogeneity = 0.30
            else:
                heterogeneity = 0.75

            extract_G = G.subgraph(i)

            lead = Leader_Identification(leader_epsilon=0.60)
            leaders = lead.leader_finding(extract_G)

            cd = Community()
            comm_list = cd.execute(extract_G, leaders, epsilon=heterogeneity, depth=2)

            for c in comm_list.keys():
                communities[comm_counter] = c
                comm_counter += 1

    cms = []
    for c in communities.values():
        cms.append(list(c))

    return cms



import networkx as nx
import timeit
import numpy as np


def __HeadTailCommunityDetection(G, finaledgelist, head_tail_ratio=0.6):

    H = nx.connected_components(G)

    for s in H:
        subgraph = nx.subgraph(G, s)
        result = nx.edge_betweenness(subgraph, normalized=False)
        edges = list(result.keys())
        values = list(result.values())
        mean = np.mean(values)
        edgelist = []
        edgetemp = subgraph.edges()
        if len(edgetemp) <= 2:
            for edge in edgetemp:
                finaledgelist.append(edge)
        else:
            for index in range(len(values)):
                if values[index] <= mean:
                    edgelist.append(edges[index])

            if (
                float(len(edgelist)) / float(len(edges)) <= head_tail_ratio
            ):  # change the head/tail division rule here, here is for tail percentage,
                # so if the rule is 40/60, the value should be assigned 0.6 as in the code.
                for edge in edgelist:
                    finaledgelist.append(edge)
            else:
                Gsub = nx.Graph()
                for edge in edgelist:
                    Gsub.add_edge(edge[0], edge[1])
                try:
                    __HeadTailCommunityDetection(Gsub, finaledgelist, head_tail_ratio)
                except:
                    pass
    return finaledgelist


def HeadTail(G):
    finaledgelist = __HeadTailCommunityDetection(G, [], head_tail_ratio=0.4)
    g1 = nx.Graph()
    g1.add_edges_from(finaledgelist)
    coms = [list(c) for c in nx.connected_components(g1)]
    return coms



import json
from collections import defaultdict
from .named_clustering import NamedClustering
import networkx as nx
from typing import Callable


class TemporalClustering(object):
    def __init__(self):
        """
        Temporal Communities representation
        """

        self.clusterings = defaultdict(NamedClustering)
        self.current_observation = 0
        self.obs_to_time = {}
        self.time_to_obs = {}
        self.matching = None
        self.matched = None

    def add_matching(self, matching: list):
        """
        Add a precomputed matching of the communities.

        :param matching: a list of tuples [(Ti_Ca, Tj_Cb, score), ... ].
                Community names needs to satisfy the pattern {tid}_{cid}, where tid is the time of observation and
                cid is the position of the community within the Clustering object.
        """
        self.matching = matching

    def get_observation_ids(self) -> list:
        """
        Returns the list of temporal ids for the available clusterings
        :return: a list of temporal ids
        """
        return list(self.time_to_obs.keys())

    def get_clustering_at(self, time: object) -> object:
        """
        Returns the clustering observed at a given time

        :param time: the time of observation
        :return: a Clustering object
        """
        return self.clusterings[self.time_to_obs[time]]

    def add_clustering(self, clustering: object, time: object):
        """
        Add to the Temporal Clustering the communities observed at a given time

        :param clustering: a Clustering object
        :param time: time of observation
        """

        named_clustering = {}
        for i, c in enumerate(clustering.communities):
            named_clustering[f"{time}_{i}"] = c

        self.clusterings[self.current_observation] = NamedClustering(
            named_clustering,
            clustering.graph,
            clustering.method_name,
            method_parameters=clustering.method_parameters,
            overlap=clustering.overlap,
        )
        self.time_to_obs[time] = self.current_observation
        self.obs_to_time[self.current_observation] = time
        self.current_observation += 1

    def get_community(self, cid: str) -> list:
        """
        Returns the nodes within a given temporal community

        :param cid: community id of the form {tid}_{cid}, where tid is the time of observation and
                cid is the position of the community within the Clustering object.
        :return: list of nodes within cid
        """

        t, cid = cid.split("_")
        coms = self.get_clustering_at(int(t))
        return coms.communities[int(cid)]

    def to_json(self):
        """
        Generate a JSON representation of the TemporalClustering object

        :return: a JSON formatted string representing the object
        """

        tcluster = {"clusters": [], "matchings": None}
        if self.matching is not None:
            tcluster["matchings"] = self.matching
        elif self.matched is not None:
            tcluster["matchings"] = self.matched

        for tid in self.get_observation_ids():
            ct = self.get_clustering_at(tid)
            partition = {
                "tid": tid,
                "communities": ct.named_communities,
                "algorithm": ct.method_name,
                "params": ct.method_parameters,
                "overlap": ct.overlap,
                "coverage": ct.node_coverage,
            }
            tcluster["clusters"].append(partition)

        return json.dumps(tcluster)

    def clustering_stability_trend(
        self, method: Callable[[object, object], float]
    ) -> list:
        """
        Returns the trend for community stability.
        The stability index is computed for temporally adjacent clustering pairs.

        :param method: a comparison score taking as input two Clustering objects (e.g., NMI, NF1, ARI...)
        :return: a list of floats
        """

        stability = []

        for i in range(self.current_observation - 1):
            c_i = self.clusterings[i]
            c_j = self.clusterings[i + 1]
            stb = method(c_i, c_j)
            stability.append(stb.score)

        return stability

    def has_explicit_match(self) -> bool:
        """
        Checks if the algorithm provided an explicit match of temporal communities

        :return: a list of tuple [(Ti_Ca, Tj_Cb, score), ... ].
                Community names are assigned following the pattern {tid}_{cid}, where tid is the time of observation and
                cid is the position of the community within the Clustering object.
        """
        if self.matching is not None:
            return True
        else:
            return False

    def get_explicit_community_match(self) -> list:
        """
        Return an explicit matching of computed communities (if it exists)

        :return: a list of tuple [(Ti_Ca, Tj_Cb, score), ... ].
                Community names are assigned following the pattern {tid}_{cid}, where tid is the time of observation and
                cid is the position of the community within the Clustering object.
        """
        return self.matching

    def community_matching(
        self, method: Callable[[set, set], float], two_sided: bool = False
    ) -> list:
        """
        Reconstruct community matches across adjacent observations using a provided similarity function.

        :param method: a set similarity function with co-domain in [0,1] (e.g., Jaccard)
        :param two_sided: boolean.
                            Whether the match has to be applied only from the past to the future (False, default)
                            or even from the future to the past (True)
        :return: a list of tuples [(Ti_Ca, Tj_Cb, score), ... ].
                Community names are assigned following the pattern {tid}_{cid}, where tid is the time of observation and
                cid is the position of the community within the Clustering object.
        """

        if self.matching is not None:
            return self.matching

        lifecycle = []

        for i in range(self.current_observation - 1):
            c_i = self.clusterings[i]
            c_j = self.clusterings[i + 1]
            for name_i, com_i in c_i.named_communities.items():

                # name_i = f"{self.obs_to_time[i]}_{cid_i}"
                best_match = []
                best_score = 0

                for name_j, com_j in c_j.named_communities.items():
                    # name_j = f"{self.obs_to_time[i+1]}_{cid_j}"

                    match = method(com_i, com_j)
                    if match > best_score:
                        best_match = [name_j]
                        best_score = match
                    elif match == best_score:
                        best_match.append(name_j)

                for j in best_match:
                    lifecycle.append((name_i, j, best_score))

        if two_sided:

            for i in range(self.current_observation - 1, 0, -1):
                c_i = self.clusterings[i]
                c_j = self.clusterings[i - 1]

                for name_i, com_i in c_i.named_communities.items():
                    # name_i = f"{self.obs_to_time[i]}_{cid_i}"
                    best_match = []
                    best_score = 0

                    for name_j, com_j in c_j.named_communities.items():
                        # name_j = f"{self.obs_to_time[i-1]}_{cid_j}"

                        match = method(com_i, com_j)
                        if match > best_score:
                            best_match = [name_j]
                            best_score = match
                        elif match == best_score:
                            best_match.append(name_j)

                    for j in best_match:
                        lifecycle.append((j, name_i, best_score))

        self.matched = lifecycle

        return lifecycle

    def lifecycle_polytree(
        self, method: Callable[[set, set], float] = None, two_sided: bool = False
    ) -> nx.DiGraph:
        """
        Reconstruct the poly-tree representing communities lifecycles using a provided similarity function.

        :param method: a set similarity function with co-domain in [0,1] (e.g., Jaccard)
        :param two_sided: boolean.
                            Whether the match has to be applied only from the past to the future (False, default)
                            or even from the future to the past (True)
        :return: a networkx DiGraph object.
                Nodes represent communities, their ids are assigned following the pattern {tid}_{cid},
                where tid is the time of observation and
                cid is the position of the community within the Clustering object.
        """

        if self.matching is not None:
            lifecycle = self.matching
        else:
            if method is None:
                raise ValueError("method parameter not specified")
            lifecycle = self.community_matching(method, two_sided)

        pt = nx.DiGraph()
        if len(lifecycle[0]) == 3:
            for u, v, w in lifecycle:
                pt.add_edge(u, v, weight=w)
        else:
            # implicit matching
            for u, v in lifecycle:
                pt.add_edge(u, v)

        return pt



import numpy as np
import scipy.sparse as sp
import networkx as nx
from collections import defaultdict
import math

__all__ = ["detect_belief_communities"]


def get_degree_vector(g: nx.Graph):
    # compute the vector of node degrees
    return sp.csr_matrix(np.array([g.degree()[i] for i in range(g.number_of_nodes())]))


def init_beliefs(q: int, g: nx.Graph):
    # initialize the beliefs

    n_n = g.number_of_nodes()
    b = sp.random(
        n_n, q, density=1, data_rvs=lambda x: np.random.random(x), format="csr"
    )
    b = sp.csr_matrix(b / b.sum(axis=1))

    return b


def init_messages(q: int, g: nx.Graph):
    # initialize the messages

    n_n = g.number_of_nodes()
    messages = sp.dok_matrix((n_n * n_n, q))

    for i in range(q):
        for node in g.nodes:
            for neighbor in g.neighbors(node):
                messages[node * n_n + neighbor, i] = np.random.random()

    sum = messages.sum(axis=1).transpose().tolist()[0]
    norm = sp.diags(sum, format="csr")
    norm[norm.nonzero()] = 1 / norm[norm.nonzero()]
    messages = messages.tocsr(copy=True)

    return norm.dot(messages)


def compute_theta(beliefs: sp.csr_matrix, degrees: sp.csr_matrix):
    # compute_theta(beliefs, degrees)[t,0] returns theta of community t
    return degrees.dot(beliefs)


def update_matrix(g: nx.Graph):
    # compute the matrix needed to update the messages

    n_n = g.number_of_nodes()
    update_mat = sp.dok_matrix((n_n * n_n, n_n * n_n), dtype=np.int8)
    for node in g.nodes:
        for neighbor in g.neighbors(node):
            for node_neighbor in g.neighbors(node):
                if node_neighbor != neighbor:
                    update_mat[node * n_n + neighbor, node_neighbor * n_n + node] = 1

    update_mat = update_mat.tocsr(copy=True)

    return update_mat


def belief_matrix(g: nx.Graph):
    # comupte the matrix needed to calculate the beliefs

    n_n = g.number_of_nodes()
    update_mat = sp.dok_matrix((n_n, n_n * n_n), dtype=np.int8)
    for node in g.nodes:
        for neighbor in g.neighbors(node):
            update_mat[node, neighbor * n_n + node] = 1

    update_mat = update_mat.tocsr(copy=True)

    return update_mat


def get_external_field_beliefs(theta: sp.csr_matrix, g: nx.Graph, beta: float):
    # compute the external field based on theta for the computation of the beliefs

    m = g.number_of_edges()

    degrees = get_degree_vector(g)

    ext_field = -(beta / (2 * m)) * degrees.transpose().dot(theta)

    return ext_field


def external_field_update_matrix(g: nx.Graph, beta: float, q: int):
    # compute the matrix needed to update the external field

    n_n = g.number_of_nodes()

    m = g.number_of_edges()

    ext_field_update_matrix = sp.dok_matrix((n_n * n_n, n_n))

    for i in range(q):
        for node in g.nodes:
            for neighbor in g.neighbors(node):
                ext_field_update_matrix[(node) * n_n + neighbor, neighbor] = -(
                    (beta * g.degree()[node]) / (2 * m)
                )

    ext_field_update_matrix = ext_field_update_matrix.tocsr()

    return ext_field_update_matrix


def get_external_field(theta: sp.csr_matrix, ext_update: sp.csr_matrix):
    # update the external field for the message update

    n_n = ext_update.shape[1]

    ones_vec = sp.csr_matrix(np.ones((n_n, 1)))

    theta_transformed = ones_vec * theta

    ext_field = ext_update * theta_transformed

    return ext_field


def run_bp_community_detection(
    g: nx.Graph,
    q: int = 2,
    beta: float = 1.012,
    max_it: int = 30,
    eps: float = 0.0001,
    reruns_if_not_conv: int = 20,
):
    # run belief propagation community detection with given parameters

    n_n = g.number_of_nodes()

    # building update_matrix, belief_update_matrix, degree_vector
    update_mat = update_matrix(g)
    belief_update_mat = belief_matrix(g)
    degrees = get_degree_vector(g)

    # building external_field_update_matrix
    ext_field_update_matrix = external_field_update_matrix(g=g, beta=beta, q=q)

    converged = False
    runs = 0
    for i in range(reruns_if_not_conv):

        if converged:
            break

        runs = runs + 1

        # initializing beliefs and messages
        beliefs = init_beliefs(q, g)
        messages = init_messages(q, g)

        # building a vector with ones at certain positions, needed for computation
        ones_vector = sp.dok_matrix((n_n * n_n, q))
        ones_vector[messages.nonzero()] = 1
        ones_vector = ones_vector.tocsr(copy=True)

        # print('Run: ', i+1)

        belief_diff = 1

        iterations = 0
        for j in range(max_it):
            iterations = iterations + 1
            if belief_diff < eps:
                converged = True
                break

            theta = compute_theta(beliefs, degrees)

            # update beliefs with updated messages
            new_messages = ones_vector + (messages * (np.exp(beta) - 1))
            new_messages.data = np.log(new_messages.data)

            updated_beliefs = belief_update_mat.dot(new_messages)
            old_beliefs = beliefs.copy()

            beliefs = updated_beliefs + get_external_field_beliefs(
                theta=theta, g=g, beta=beta
            )
            beliefs.data = np.exp(beliefs.data)

            b_sum = beliefs.sum(axis=1).transpose().tolist()[0]
            b_norm = sp.diags(b_sum, format="csr")
            b_norm.data = 1 / b_norm.data
            beliefs = b_norm.dot(beliefs)
            beliefs = beliefs.tocsr(copy=True)

            # update messages

            external_field = get_external_field(
                theta=theta, ext_update=ext_field_update_matrix
            )

            updated_messages = update_mat.dot(new_messages)

            messages = external_field + updated_messages

            messages.data = np.exp(messages.data)

            m_sum = messages.sum(axis=1).transpose().tolist()[0]
            m_norm = sp.diags(m_sum, format="csr")
            m_norm.data = 1 / m_norm.data
            messages = m_norm.dot(messages)
            messages = messages.tocsr(copy=True)

            belief_diff = np.linalg.norm(beliefs.todense() - old_beliefs)

    detected_communities = [np.argmax(beliefs[i, :]) for i in range(beliefs.shape[0])]

    return beliefs, detected_communities, runs, iterations, converged


def compute_opt_beta(q: int, c: int):
    # compute the optimal value for the parameter beta

    return math.log(q / (math.sqrt(c) - 1) + 1)


def detect_belief_communities(
    g: nx.Graph,
    max_it: int = 100,
    eps: float = 0.0001,
    reruns_if_not_conv: int = 5,
    threshold: float = 0.005,
    q_max: int = 7,
):

    from networkx.algorithms.community.quality import modularity

    # determine number of optimal communities and run community detection for a given network
    # The nodes have to be labeled form 0 to n

    modularity_0 = 0
    modularity_1 = threshold
    q = 1
    c = 2 * g.number_of_edges() / g.number_of_nodes()
    partition = ()

    # run belief propagation community detection with increasing number of communities until the modularity of the
    # detected partition does not increase more then given threshold

    while modularity_1 - modularity_0 >= threshold:
        old_partition = partition
        beta = compute_opt_beta(q, c)
        modularity_0 = modularity_1
        partition = run_bp_community_detection(
            g=g,
            q=q,
            beta=beta,
            max_it=max_it,
            eps=eps,
            reruns_if_not_conv=reruns_if_not_conv,
        )

        modularity_1 = modularity(
            g,
            [
                {i for i in range(len(partition[1])) if partition[1][i] == j}
                for j in set(partition[1])
            ],
        )

        if not partition[4]:
            curr_partition = partition
            partition = old_partition
            modularity_1 = modularity_0
            modularity_0 = modularity_1 - threshold

        if q == 1:
            modularity_0 = modularity_1 - threshold

        if q > q_max:
            break

        q = q + 1

    if len(old_partition) != 0:
        return prepare_coms(g, old_partition[1])
    else:
        return prepare_coms(g, curr_partition[1])


def prepare_coms(g, cms):
    ass = list(zip(g.nodes(), cms))

    coms = defaultdict(list)
    for i in ass:
        coms[i[1]].append(i[0])

    return list(coms.values())



from collections import namedtuple
import itertools
from random import sample

import networkx as nx
import numpy as np
import cdlib
from typing import Callable

__all__ = [
    "BoolParameter",
    "Parameter",
    "grid_execution",
    "grid_search",
    "pool",
    "pool_grid_filter",
    "random_search",
]

Parameter = namedtuple("Parameter", "name start end step")
Parameter.__new__.__defaults__ = (None,) * len(Parameter._fields)

BoolParameter = namedtuple("BoolParameter", "name value")
BoolParameter.__new__.__defaults__ = (None,) * len(BoolParameter._fields)


def __generate_ranges(parameter: tuple) -> list:
    """

    :param parameter:
    :return:
    """
    values = []
    if isinstance(parameter, cdlib.ensemble.Parameter):
        if parameter.step is None:
            values.append((parameter.name, parameter.start))
        else:
            for actual in np.arange(parameter.start, parameter.end, parameter.step):
                if isinstance(actual, np.int64):
                    actual = int(actual)
                values.append((parameter.name, actual))

    elif isinstance(parameter, BoolParameter):
        if parameter.value is None:
            values = [(parameter.name, True), (parameter.name, False)]
        else:
            values = [(parameter.name, parameter.value)]
    else:
        raise ValueError(
            "parameter should be either an instance of Parameter or of BoolParameter"
        )
    return values


def grid_execution(
    graph: nx.Graph, method: Callable[[nx.Graph, dict], object], parameters: list
) -> tuple:
    """
    Instantiate the specified community discovery method performing a grid search on the parameter set.

    :param method: community discovery method (from nclib.community)
    :param graph: networkx/igraph object
    :param parameters: list of Parameter and BoolParameter objects
    :return: at each call the generator yields a tuple composed by the current configuration and the obtained communities

    :Example:

    >>> import networkx as nx
    >>> from cdlib import algorithms, ensemble
    >>> g = nx.karate_club_graph()
    >>> resolution = ensemble.Parameter(name="resolution", start=0.1, end=1, step=0.1)
    >>> for communities in ensemble.grid_execution(graph=g, method=algorithms.louvain, parameters=[resolution]):
    >>>     print(communities)
    """
    configurations = []
    for parameter in parameters:
        configurations.append(__generate_ranges(parameter))

    for element in itertools.product(*configurations):
        values = {e[0]: e[1] for e in element}
        res = method(graph, **values)
        yield res


def grid_search(
    graph: nx.Graph,
    method: Callable[[nx.Graph, dict], object],
    parameters: list,
    quality_score: Callable[[nx.Graph, object], object],
    aggregate: Callable[[list], object] = max,
) -> tuple:
    """
    Returns the optimal partition of the specified graph w.r.t. the selected algorithm and quality score.

    :param method: community discovery method (from nclib.community)
    :param graph: networkx/igraph object
    :param parameters: list of Parameter and BoolParameter objects
    :param quality_score: a fitness function to evaluate the obtained partition (from nclib.evaluation)
    :param aggregate: function to select the best fitness value. Possible values: min/max
    :return: at each call the generator yields a tuple composed by: the optimal configuration for the given algorithm, input paramters and fitness function; the obtained communities; the fitness score

    :Example:

    >>> import networkx as nx
    >>> from cdlib import algorithms, ensemble
    >>> g = nx.karate_club_graph()
    >>> resolution = ensemble.Parameter(name="resolution", start=0.1, end=1, step=0.1)
    >>> randomize = ensemble.BoolParameter(name="randomize")
    >>> communities, scoring = ensemble.grid_search(graph=g, method=algorithms.louvain,
    >>>                                                     parameters=[resolution, randomize],
    >>>                                                     quality_score=evaluation.erdos_renyi_modularity,
    >>>                                                     aggregate=max)
    >>> print(communities, scoring)
    """
    results = {}
    for communities in grid_execution(graph, method, parameters):
        results[tuple(communities.method_parameters.items())] = {
            "communities": communities,
            "scoring": quality_score(graph, communities),
        }

    res = aggregate(results, key=lambda x: results.get(x)["scoring"])

    return results[res]["communities"], results[res]["scoring"]


def random_search(
    graph: nx.Graph,
    method: Callable[[nx.Graph, dict], object],
    parameters: list,
    quality_score: Callable[[nx.Graph, object], object],
    instances: int = 10,
    aggregate: Callable[[list], object] = max,
) -> tuple:
    """
    Returns the optimal partition of the specified graph w.r.t. the selected algorithm and quality score over a randomized sample of the input parameters.

    :param method: community discovery method (from nclib.community)
    :param graph: networkx/igraph object
    :param parameters: list of Parameter and BoolParameter objects
    :param quality_score: a fitness function to evaluate the obtained partition (from nclib.evaluation)
    :param instances: number of randomly selected parameters configurations
    :param aggregate: function to select the best fitness value. Possible values: min/max

    :return: at each call the generator yields a tuple composed by: the optimal configuration for the given algorithm, input paramters and fitness function; the obtained communities; the fitness score

    :Example:

    >>> import networkx as nx
    >>> from cdlib import algorithms, ensemble
    >>> g = nx.karate_club_graph()
    >>> resolution = ensemble.Parameter(name="resolution", start=0.1, end=1, step=0.1)
    >>> randomize = ensemble.BoolParameter(name="randomize")
    >>> communities, scoring = ensemble.random_search(graph=g, method=algorithms.louvain,
    >>>                                                       parameters=[resolution, randomize],
    >>>                                                       quality_score=evaluation.erdos_renyi_modularity,
    >>>                                                       instances=5, aggregate=max)
    >>> print(communities, scoring)
    """

    configurations = []
    for parameter in parameters:
        configurations.append(__generate_ranges(parameter))

    # configuration sampling
    selected = sample(list(itertools.product(*configurations)), instances)

    results = {}
    for element in selected:
        values = {e[0]: e[1] for e in element}
        communities = method(graph, **values)
        results[element] = {
            "communities": communities,
            "scoring": quality_score(graph, communities),
        }

    res = aggregate(results, key=lambda x: results.get(x)["scoring"])

    return results[res]["communities"], results[res]["scoring"]


def pool(
    graph: nx.Graph, methods: Callable[[nx.Graph, dict], object], configurations: list
) -> tuple:
    """
    Execute on a pool of community discovery internal on the input graph.

    :param methods: list community discovery methods (from nclib.community)
    :param graph: networkx/igraph object
    :param configurations: list of lists (one for each method) of Parameter and BoolParameter objects
    :return: at each call the generator yields a tuple composed by: the actual method, its current configuration and the obtained communities
    :raises ValueError: if the number of methods is different from the number of configurations specified

    :Example:

    >>> import networkx as nx
    >>> from cdlib import algorithms, ensemble
    >>> g = nx.karate_club_graph()
    >>> # Louvain
    >>> resolution = ensemble.Parameter(name="resolution", start=0.1, end=1, step=0.1)
    >>> randomize = ensemble.BoolParameter(name="randomize")
    >>> louvain_conf = [resolution, randomize]
    >>>
    >>> # Angel
    >>> threshold = ensemble.Parameter(name="threshold", start=0.1, end=1, step=0.1)
    >>> angel_conf = [threshold]
    >>>
    >>> methods = [algorithms.louvain, algorithms.angel]
    >>>
    >>> for communities in ensemble.pool(g, methods, [louvain_conf, angel_conf]):
    >>>     print(communities)
    """
    if len(methods) != len(configurations):
        raise ValueError("The number of methods and configurations must match")

    for i in range(len(methods)):
        for res in grid_execution(graph, methods[i], configurations[i]):
            yield res


def pool_grid_filter(
    graph: nx.Graph,
    methods: Callable[[nx.Graph, dict], object],
    configurations: list,
    quality_score: Callable[[nx.Graph, object], object],
    aggregate: Callable[[list], object] = max,
) -> tuple:
    """
    Execute a pool of community discovery internal on the input graph.
    Returns the optimal partition for each algorithm given the specified quality function.

    :param methods: list community discovery methods (from nclib.community)
    :param graph: networkx/igraph object
    :param configurations: list of lists (one for each method) of Parameter and BoolParameter objects
    :param quality_score: a fitness function to evaluate the obtained partition (from nclib.evaluation)
    :param aggregate: function to select the best fitness value. Possible values: min/max
    :return: at each call the generator yields a tuple composed by: the actual method, its optimal configuration; the obtained communities; the fitness score.
    :raises ValueError: if the number of methods is different from the number of configurations specified

    :Example:

    >>> import networkx as nx
    >>> from cdlib import algorithms, ensemble
    >>> g = nx.karate_club_graph()
    >>> # Louvain
    >>> resolution = ensemble.Parameter(name="resolution", start=0.1, end=1, step=0.1)
    >>> randomize = ensemble.BoolParameter(name="randomize")
    >>> louvain_conf = [resolution, randomize]
    >>>
    >>> # Angel
    >>> threshold = ensemble.Parameter(name="threshold", start=0.1, end=1, step=0.1)
    >>> angel_conf = [threshold]
    >>>
    >>> methods = [algorithms.louvain, algorithms.angel]
    >>>
    >>> for communities, scoring in ensemble.pool_grid_filter(g, methods, [louvain_conf, angel_conf], quality_score=evaluation.erdos_renyi_modularity, aggregate=max):
    >>>     print(communities, scoring)

    """
    if len(methods) != len(configurations):
        raise ValueError("The number of methods and configurations must match")

    for i in range(len(methods)):
        communities, scoring = grid_search(
            graph, methods[i], configurations[i], quality_score, aggregate
        )
        yield communities, scoring
